# test module for the derived RelPerm class located in resqpy.olio.dataframe.py

import os
import pytest

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

import resqpy.model as rq
from resqpy.olio.relperm import RelPerm, text_to_relperm_dict, relperm_parts_in_model


def test_col_headers(tmp_path):
   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.683, np.nan],
                     [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.018, np.nan], [0.74, 1.0, 0.0, 1e-06]])
   df_cols1 = ['Sw', 'Krg', 'Kro', 'Pc']
   df_cols2 = ['Sg', 'Krg', 'Krw', 'Pc']
   df_cols3 = ['Sg', 'Krg', 'Pc', 'Kro']
   df_cols4 = ['Sg', 'krg', 'Kro', 'pC']
   df1 = pd.DataFrame(np_df, columns = df_cols1)
   df2 = pd.DataFrame(np_df, columns = df_cols2)
   df3 = pd.DataFrame(np_df, columns = df_cols3)
   df4 = pd.DataFrame(np_df, columns = df_cols4)
   phase_combo1 = 'gas-oil'
   phase_combo2 = None

   with pytest.raises(ValueError) as excval1:
      RelPerm(model = model, df = df1, phase_combo = phase_combo1)
   assert "incorrect saturation column name and/or multiple saturation columns exist" in str(excval1.value)

   with pytest.raises(ValueError) as excval2:
      RelPerm(model = model, df = df2, phase_combo = phase_combo1)
   assert "incorrect column name(s) {'Krw'}" in str(excval2.value)

   with pytest.raises(ValueError) as excval3:
      RelPerm(model = model, df = df3, phase_combo = phase_combo1)
   assert "capillary pressure data should be in the last column of the dataframe" in str(excval3.value)

   relperm_obj = RelPerm(model = model, df = df4, phase_combo = phase_combo2)
   assert relperm_obj.phase_combo == 'gas-oil'


def test_missing_vals(tmp_path):
   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, np.nan, 0.683, np.nan],
                     [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.018, np.nan], [0.74, 1.0, np.nan, 1e-06]])
   df_cols = ['Sg', 'Krg', 'Kro', 'Pc']
   df = pd.DataFrame(np_df, columns = df_cols)
   phase_combo = 'gas-oil'

   with pytest.raises(Exception) as excval:
      RelPerm(model = model, df = df, phase_combo = phase_combo)
   assert "missing values found in Krg column" in str(excval.value)


def test_monotonicity(tmp_path):
   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.683, np.nan],
                     [0.25, 0.205, 0.35, np.nan], [0.85, 0.55, 0.018, 1e-06], [0.74, 1.0, 0.0, 1e-07]])
   df_cols = ['Sg', 'Krg', 'Kro', 'Pc']
   df = pd.DataFrame(np_df, columns = df_cols)
   phase_combo = 'gas-oil'

   with pytest.raises(ValueError) as excval:
      RelPerm(model = model, df = df, phase_combo = phase_combo)
   assert "('Sg', 'Krg', 'Kro') combo is not monotonic" in str(excval.value)

   df['Sg'] = [0.0, 0.04, 0.12, 0.25, 0.45, 0.74]
   with pytest.raises(ValueError) as excval1:
      RelPerm(model = model, df = df, phase_combo = phase_combo)
   assert "Pc values are not monotonic" in str(excval1.value)


def test_range(tmp_path):
   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.683, np.nan],
                     [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.018, np.nan], [0.74, 1.1, -0.001, 1e-06]])
   df_cols = ['Sg', 'Krg', 'Kro', 'Pc']
   df = pd.DataFrame(np_df, columns = df_cols)
   phase_combo = 'gas-oil'

   with pytest.raises(ValueError) as excval:
      RelPerm(model = model, df = df, phase_combo = phase_combo)
   assert "Krg is not within the range 0-1" in str(excval.value)


def test_relperm(tmp_path):
   epc = os.path.join(tmp_path, 'model.epc')
   model = rq.new_model(epc)
   np_df = np.array([[0.0, 0.0, 1.0, 0], [0.04, 0.015, 0.87, np.nan], [0.12, 0.065, 0.689, np.nan],
                     [0.25, 0.205, 0.35, np.nan], [0.45, 0.55, 0.019, np.nan], [0.74, 1.0, 0.0, 0.000001]])
   df_cols1 = ['Sw', 'Krw', 'Kro', 'Pc']
   df1 = pd.DataFrame(np_df, columns = df_cols1)
   phase_combo1 = 'oil-water'
   uom_list = ['Euc', 'Euc', 'Euc', 'psi']
   df_cols2 = ['Sg', 'Krg', 'Kro', 'Pc']
   df2 = pd.DataFrame(np_df, columns = df_cols2)
   phase_combo2 = 'gas-oil'
   dataframe1 = RelPerm(model = model,
                        df = df1,
                        uom_list = uom_list,
                        phase_combo = phase_combo1,
                        low_sal = True,
                        table_index = 1,
                        title = 'table1')
   dataframe2 = RelPerm(model = model,
                        df = df2,
                        uom_list = uom_list,
                        phase_combo = phase_combo2,
                        low_sal = False,
                        title = 'table2')
   assert dataframe1.n_cols == 4
   assert dataframe1.n_rows == 6
   assert_frame_equal(dataframe1.dataframe(), df1)
   assert all(dataframe1.dataframe().columns == df_cols1)
   assert round(dataframe1.interpolate_point(saturation = 0.55, kr_or_pc_col = 'Kro')[1], 3) == 0.012
   dataframe1.write_hdf5_and_create_xml()
   dataframe2.write_hdf5_and_create_xml()
   assert model.part(extra = {'relperm_table': 'true', 'low_sal': 'true'}) == model.part(title = 'table1')
   assert len(relperm_parts_in_model(model, low_sal = False)) == 1
   dataframe1.df_to_text(filepath = tmp_path, filename = 'oil_water_test_table')
   df1_reconstructed = text_to_relperm_dict(os.path.join(tmp_path, 'oil_water_test_table.dat'))['relperm_table1']['df']
   assert_frame_equal(df1, df1_reconstructed)
   assert df1_reconstructed.iloc[3]['Kro'] == 0.350
   # initialize a RelPerm object from an existing uuid
   new_wo_obj = RelPerm(model, uuid = model.uuid(title = 'table1'))
   assert new_wo_obj.phase_combo == phase_combo1
   assert new_wo_obj.low_sal == 'true'
   assert new_wo_obj.table_index == '1'
   assert new_wo_obj.extra_metadata == {
      'dataframe': 'true',
      'low_sal': 'true',
      'phase_combo': 'oil-water',
      'relperm_table': 'true',
      'table_index': '1'
   }



"""polylines.py: Resqml polylines module."""

version = '14th July 2021'

import logging

log = logging.getLogger(__name__)
log.debug('polylines.py version ' + version)

from resqpy.olio.base import BaseResqpy
import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vu
import resqpy.olio.intersection as meet
import resqpy.olio.point_inclusion as pip
from resqpy.olio.xml_namespaces import curly_namespace as ns
import resqpy.olio.write_hdf5 as rwh5
import resqpy.crs as rcrs
import resqpy.olio.simple_lines as rsl
import resqpy.organize as rqo

import math as maths
import numpy as np
import os


class _BasePolyline(BaseResqpy):
   """Base class to implement shared methods for other classes in this module"""

   def create_interpretation_and_feature(self,
                                         kind = 'horizon',
                                         name = None,
                                         interp_title_suffix = None,
                                         is_normal = True):
      """Creates xml and objects for a represented interpretaion and interpreted feature, if not already present."""

      assert kind in ['horizon', 'fault', 'fracture', 'geobody boundary']
      assert name or self.title, 'title missing'
      if not name:
         name = self.title

      if self.rep_int_root is not None:
         log.debug(f'represented interpretation already exisrs for surface {self.title}')
         return
      if kind in ['horizon', 'geobody boundary']:
         feature = rqo.GeneticBoundaryFeature(self.model, kind = kind, feature_name = name)
         feature.create_xml()
         if kind == 'horizon':
            interp = rqo.HorizonInterpretation(self.model, genetic_boundary_feature = feature, domain = 'depth')
         else:
            interp = rqo.GeobodyBoundaryInterpretation(self.model, genetic_boundary_feature = feature, domain = 'depth')
      elif kind in ['fault', 'fracture']:
         feature = rqo.TectonicBoundaryFeature(self.model, kind = kind, feature_name = name)
         feature.create_xml()
         interp = rqo.FaultInterpretation(self.model,
                                          is_normal = is_normal,
                                          tectonic_boundary_feature = feature,
                                          domain = 'depth')  # might need more arguments
      else:
         log.critical('code failure')
      interp_root = interp.create_xml(title_suffix = interp_title_suffix)
      self.rep_int_root = interp_root


def load_hdf5_array(object, node, array_attribute, tag = 'Values'):
   """Loads the property array data as an attribute of object, from the hdf5 referenced in xml node.

   :meta private:
   """

   assert (rqet.node_type(node) in ['DoubleHdf5Array', 'IntegerHdf5Array', 'Point3dHdf5Array'])
   # ignore null value
   h5_key_pair = object.model.h5_uuid_and_path_for_node(node, tag = tag)
   if h5_key_pair is None:
      return None
   return object.model.h5_array_element(h5_key_pair,
                                        index = None,
                                        cache_array = True,
                                        object = object,
                                        array_attribute = array_attribute)


class Polyline(_BasePolyline):
   """Class for RESQML polyline representation."""

   resqml_type = 'PolylineRepresentation'

   def __init__(self,
                parent_model,
                poly_root = None,
                uuid = None,
                set_bool = None,
                set_coord = None,
                set_crs = None,
                set_crsroot = None,
                title = None,
                rep_int_root = None,
                originator = None,
                extra_metadata = None):
      """Initialises a new PolylineRepresentation object.

        arguments:
            parent_model (model.Model object): the model which the new PolylineRepresentation belongs to
            poly_root (DEPRECATED): use uuid instead;
                the root node of the xml tree representing the PolylineRepresentation;
                if not None, the new PolylineRepresentation object is initialised based on data in tree;
                if None, wait for further improvements
            uuid (uuid.UUID, optional): the uuid of an existing RESQML PolylineRepresentation from which
                to initialise the resqpy Polyline
            set_bool (boolean, optional): if True, a new polyline created from coordinates is flagged as
                a closed polyline (polygon); ignored if uuid or poly_root is not None
            set_coord (numpy array of shape (..., 3), optional): an ordered set of xyz values used to define
                a new polyline; ignored if uuid or poly_root is not None
            set_crs (uuid.UUID, optional): the uuid of a crs to be used when initialising from coordinates;
                ignored if uuid or poly_root is not None
            set_crsroot (DEPRECATED): the xml root node for the crs; ignored
            title (str, optional): the citation title to use for a new polyline;
                ignored if uuid or poly_root is not None
            rep_int_root
            originator (str, optional): the name of the person creating the polyline, defaults to login id;
                ignored if uuid or poly_root is not None
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the polyline;
                ignored if uuid or poly_root is not None

        returns:
            the newly instantiated Polyline object

        :meta common:
        """

      self.model = parent_model
      self.isclosed = set_bool
      self.nodepatch = None
      self.crs_uuid = set_crs
      self.coordinates = None
      self.centre = None
      self.rep_int_root = rep_int_root  # Optional
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       originator = originator,
                       root_node = poly_root,
                       extra_metadata = extra_metadata)

      if self.root is None and all(i is not None for i in [set_bool, set_coord, set_crs, title]):
         # Using data from a polyline set
         assert set_coord.ndim > 1 and 2 <= set_coord.shape[-1] <= 3
         if rep_int_root is not None:
            self.rep_int_root = rep_int_root
         else:
            self.rep_int_root = None
         # allow for x,y or x,y,z incoming coordinates but use x,y,z internally
         coord_shape = list(set_coord.shape)
         coord_shape[-1] = 3
         self.coordinates = np.zeros(tuple(coord_shape))
         self.coordinates[..., :set_coord.shape[-1]] = set_coord
         if set_coord.ndim > 2:
            self.coordinates = self.coordinates.reshape((-1, 3))
         self.nodepatch = (0, len(self.coordinates))
         assert not any(map(lambda x: x is None, self.nodepatch))  # Required fields - assert neither are None

      # TODO: Add SeismicCoordinates later - optional field
      # TODO: Add LineRole later - optional field

   def _load_from_xml(self):

      assert self.root is not None  # polyline xml node is specified
      poly_root = self.root

      self.title = rqet.citation_title_for_node(poly_root)

      self.extra_metadata = rqet.load_metadata_from_xml(self.root)

      self.isclosed = rqet.bool_from_text(rqet.node_text(rqet.find_tag(poly_root, 'IsClosed')))
      assert self.isclosed is not None  # Required field

      patch_node = rqet.find_tag(poly_root, 'NodePatch')
      assert patch_node is not None  # Required field

      geometry_node = rqet.find_tag(patch_node, 'Geometry')
      assert geometry_node is not None  # Required field

      self.crs_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(geometry_node, ['LocalCrs', 'UUID']))
      assert self.crs_uuid is not None  # Required field

      points_node = rqet.find_tag(geometry_node, 'Points')
      assert points_node is not None  # Required field
      load_hdf5_array(self, points_node, 'coordinates', tag = 'Coordinates')

      self.nodepatch = (rqet.find_tag_int(patch_node, 'PatchIndex'), rqet.find_tag_int(patch_node, 'Count'))
      assert not any(map(lambda x: x is None, self.nodepatch))  # Required fields - assert neither are None

      self.rep_int_root = self.model.referenced_node(rqet.find_tag(poly_root, 'RepresentedInterpretation'))

   @property
   def crs_root(self):
      """XML node corresponding to self.crs_uuid"""

      return self.model.root_for_uuid(self.crs_uuid)

   @property
   def rep_int_uuid(self):
      # TODO: Track uuid only, not root
      return rqet.uuid_for_part_root(self.rep_int_root)

   def is_convex(self, trust_metadata = True):
      """Returns True if the polyline is closed and convex in the xy plane, otherwise False."""

      if not self.isclosed:
         return False
      if trust_metadata and self.extra_metadata is not None and 'is_convex' in self.extra_metadata.keys():
         return self.extra_metadata['is_convex']
      nodes = len(self.coordinates)
      extras = {}
      if nodes < 3:
         result = False
      else:
         cw_bool = None
         result = True
         for node in range(3, nodes):
            cw = vu.clockwise(self.coordinates[node - 2], self.coordinates[node - 1], self.coordinates[node])
            if cw == 0.0:
               continue  # striaght line section
            if cw_bool is None:
               cw_bool = (cw > 0.0)
            elif cw_bool != (cw > 0.0):
               result = False
               break
         if cw_bool is None:
            result = False  # whole polyline lies in a straight line
         if result:
            extras['is_clockwise'] = cw_bool
      extras['is_convex'] = result
      self.append_extra_metadata(extras)
      return result

   def point_is_inside_xy(self, p, mode = 'crossing'):
      """Returns True if point p is inside closed polygon, in xy plane, otherwise False.

        :meta common:
        """

      assert mode in ['crossing', 'winding'], 'unrecognised point inclusion mode: ' + str(mode)
      assert self.isclosed, 'point inclusion is not applicable to unclosed polylines'

      if mode == 'crossing':
         return pip.pip_cn(p, self.coordinates)
      return pip.pip_wn(p, self.coordinates)

   def segment_length(self, segment_index, in_xy = False):
      """Returns the naive length (ie. assuming x,y & z units are the same) of an individual segment of the polyline."""

      successor = segment_index + 1
      if self.isclosed:
         assert 0 <= segment_index < len(self.coordinates)
         if segment_index == len(self.coordinates) - 1:
            successor = 0
      else:
         assert 0 <= segment_index < len(self.coordinates) - 1

      d = 2 if in_xy else 3
      return vu.naive_length(self.coordinates[successor, :d] - self.coordinates[segment_index, :d])

   def full_length(self, in_xy = False):
      """Returns the naive length of the entire polyline.

        :meta common:
        """

      length = 0.0
      end_index = len(self.coordinates) - 1
      if self.isclosed:
         end_index += 1
      for i in range(end_index):
         length += self.segment_length(i, in_xy = in_xy)
      return length

   def interpolated_point(self, fraction, in_xy = False):
      """Returns x,y,z point on the polyline at fractional distance along entire polyline.

        :meta common:
        """

      assert 0.0 <= fraction <= 1.0
      target = fraction * self.full_length(in_xy = in_xy)
      seg_index = 0
      while True:
         next_seg_length = self.segment_length(seg_index, in_xy = in_xy)
         if target > next_seg_length or next_seg_length == 0.0:
            target -= next_seg_length
            seg_index += 1
            continue
         seg_fraction = target / next_seg_length
         successor = (seg_index + 1) % len(self.coordinates)
         return seg_fraction * self.coordinates[successor] + (1.0 - seg_fraction) * self.coordinates[seg_index]

   def equidistant_points(self, n, in_xy = False):
      """Returns array of shape (n, 3) being points equally distributed along entire polyline."""

      assert n > 1
      sample_points = np.empty((n, 3))
      if self.isclosed:
         step_fraction = 1.0 / float(n)
      else:
         step_fraction = 1.0 / float(n - 1)
      f = 0.0
      for step in range(n):
         sample_points[step] = self.interpolated_point(f, in_xy = in_xy)
         f += step_fraction
         if f > 1.0:
            f = 1.0  # to avoid possible assertion error due to rounding
      return sample_points

   def balanced_centre(self, mode = 'weighted', n = 20, cache = True, in_xy = False):
      """Returns a mean x,y,z based on sampling polyline at regular intervals."""

      if cache and self.centre is not None:
         return self.centre
      assert mode in ['weighted', 'sampled']
      if mode == 'sampled':  # this mode is deprecated as it simply approximates the weighted mode
         sample_points = self.equidistant_points(n, in_xy = in_xy)
         centre = np.mean(sample_points, axis = 0)
      else:  # 'weighted'
         sum = np.zeros(3)
         seg_count = len(self.coordinates) - 1
         if self.isclosed:
            seg_count += 1
         d = 2 if in_xy else 3
         for seg_index in range(seg_count):
            successor = (seg_index + 1) % len(self.coordinates)
            p1, p2 = self.coordinates[seg_index, :d], self.coordinates[successor, :d]
            sum += (p1 + p2) * vu.naive_length(p2 - p1)
         centre = sum / (2.0 * self.full_length(in_xy = in_xy))
      if cache:
         self.centre = centre
      return centre

   def first_line_intersection(self, x1, y1, x2, y2, half_segment = False):
      """Returns segment number & x, y of first intersection of (half) bounded line x,y 1 to 2 with polyline, or None, None, None."""

      seg_count = len(self.coordinates) - 1
      if self.isclosed:
         seg_count += 1

      for segment in range(seg_count):
         successor = (segment + 1) % len(self.coordinates)
         x, y = meet.line_line_intersect(x1,
                                         y1,
                                         x2,
                                         y2,
                                         self.coordinates[segment][0],
                                         self.coordinates[segment][1],
                                         self.coordinates[successor][0],
                                         self.coordinates[successor][1],
                                         line_segment = True,
                                         half_segment = half_segment)
         if x is not None:
            return segment, x, y
      return None, None, None

   def normalised_xy(self, x, y, mode = 'square'):
      """Returns a normalised x',y' pair (in range 0..1) being point x,y under mapping from convex polygon.

        arguments:
            x, y (floats): location of a point inside the polyline, which must be closed and project to a
                convex polygon in the xy plane
            mode (string): which mapping algorithm to use, one of: 'square', 'circle', or 'perimeter'

        returns:
            x', y' (floats, each in range 0..1) being the normalised representation of point x,y

        notes:
            this method is the inverse of denormalised_xy(), as long as a consistent mode is selected;
            for more details of the mapping used by the 3 modes, see documentation for denormalised_xy()
        """

      assert mode in ['square', 'perimeter', 'circle']
      assert self.is_convex(), 'attempt to find normalised x,y within a polyline that is not a closed convex polygon'
      centre_xy = self.balanced_centre()[:2]
      if tuple(centre_xy) == (x, y):
         if mode == 'square':
            return 0.5, 0.5
         return 0.0, 0.0
      segment, px, py = self.first_line_intersection(centre_xy[0], centre_xy[1], x, y, half_segment = True)
      assert px is not None
      norm_x, norm_y = None, None
      if mode == 'square':
         if px == centre_xy[0]:
            norm_x = 0.5
         else:
            norm_x = 0.5 + 0.5 * (x - centre_xy[0]) / abs(px - centre_xy[0])
         if py == centre_xy[1]:
            norm_y = 0.5
         else:
            norm_y = 0.5 + 0.5 * (y - centre_xy[1]) / abs(py - centre_xy[1])
      elif mode == 'circle':
         d_xy = np.array((x, y)) - centre_xy
         dp_xy = np.array((px, py)) - centre_xy
         if abs(dp_xy[0]) > abs(dp_xy[1]):
            rf = d_xy[0] / dp_xy[0]
            theta = maths.atan(d_xy[1] / d_xy[0])
            if d_xy[0] < 0.0:
               theta += maths.pi
         else:
            rf = d_xy[1] / dp_xy[1]
            theta = maths.pi / 2.0 - maths.atan(d_xy[0] / d_xy[1])
            if d_xy[1] < 0.0:
               theta += maths.pi
         norm_x = rf * rf
         theta /= 2.0 * maths.pi
         if theta < 0.0:
            theta += 1.0
         elif theta > 1.0:
            theta -= 1.0
         norm_y = theta
      elif mode == 'perimeter':
         d_xy = np.array((x, y)) - centre_xy
         dp_xy = np.array((px, py)) - centre_xy
         if abs(dp_xy[0]) > abs(dp_xy[1]):
            rf = d_xy[0] / dp_xy[0]
         else:
            rf = d_xy[1] / dp_xy[1]
         norm_x = rf * rf
         seg_xy = np.array((px, py)) - self.coordinates[segment, :2]
         length = 0.0
         for seg in range(segment):
            length += self.segment_length(seg, in_xy = True)
         length += vu.naive_length(seg_xy)
         norm_y = length / self.full_length(in_xy = True)
      return norm_x, norm_y

   def denormalised_xy(self, norm_x, norm_y, mode = 'perimeter'):
      """Returns a denormalised x,y pair being point norm_x,norm_y (in range 0..1) under mapping onto convex polygon.

        arguments:
            norm_x, norm_y (floats): normalised values, each in range 0..1, identifying a location in a unit shape
            mode (string): 'square', 'circle', or 'perimeter'; if square, norm_x and norm_y are coordinates within a
                unit square; if circle or perimeter, norm_x is the square of a radial fraction between the centre
                of the polygon and the perimeter; if mode is circle, norm_y is simply the polar coordinate bearing
                of the point relative to the polygon centre, with values 0..1 corresponding to 0..2pi radians;
                if mode is perimeter, norm_y is a fractional distance along the length of the closed polyline
                projected onto the xy plane

        returns:
            x, y (floats): the location of a point within the area outlined by the polyline, which must be closed
                and project to a convex polygon in the xy plane

        notes:
            this method is the inverse of normalised_xy(), as long as a consistent mode is selected;
            each mode gives a reversible mapping but there are some variations in density which could introduce
            slightly different biases when working with stochastic locations; density distortion with square and
            circle modes tends to increase as the polygon shape becomes less square (and aligned) or circular
            respectively; perimeter mode is believed to yield least density distortion generally;
            for circle and perimeter modes, the normalised x value is treated as the square of the fractional
            distance to the boundary in order for a rectangular distribution of normalised values to map to
            an even density of denormalised locations
        """

      assert mode in ['square', 'perimeter', 'circle']
      assert 0.0 <= norm_x <= 1.0 and 0.0 <= norm_y <= 1.0
      assert self.is_convex(), 'attempt to find denormalised x,y within a polyline that is not a closed convex polygon'
      centre_xy = self.balanced_centre()[:2]
      x, y = None, None
      if mode == 'square':
         if (norm_x, norm_y) == (0.5, 0.5):
            return tuple(centre_xy)
         n_x = (norm_x - 0.5) + centre_xy[0]
         n_y = (norm_y - 0.5) + centre_xy[1]
         _, px, py = self.first_line_intersection(centre_xy[0], centre_xy[1], n_x, n_y, half_segment = True)
         assert px is not None
         fx = abs(norm_x - 0.5)
         fy = abs(norm_y - 0.5)
         f = 2.0 * max(fx, fy)
         log.debug(f'intersect x,y: {px}, {py}')
         if px == centre_xy[0]:
            x = centre_xy[0]
         else:
            x = centre_xy[0] + f * (px - centre_xy[0])
         if py == centre_xy[1]:
            y = centre_xy[1]
         else:
            y = centre_xy[1] + f * (py - centre_xy[1])
         log.debug(f'denormal x,y: {x}, {y}')
      elif mode == 'perimeter':
         if norm_y == 1.0:
            norm_y = 0.0
         p_xy = np.empty(2)
         p_xy[:] = self.interpolated_point(norm_y)[:2]
         f = maths.sqrt(norm_x)
         x, y = centre_xy + f * (p_xy - centre_xy)
      elif mode == 'circle':
         theta = 2 * maths.pi * norm_y
         n_x = centre_xy[0] + maths.cos(theta)
         n_y = centre_xy[1] + maths.sin(theta)
         p_xy = np.empty(2)
         p_xy[:] = self.first_line_intersection(centre_xy[0], centre_xy[1], n_x, n_y, half_segment = True)[1:]
         f = maths.sqrt(norm_x)
         x, y = centre_xy + f * (p_xy - centre_xy)
      return x, y

   def tangent_vectors(self):
      """Returns a numpy array of unit length tangent vectors, one for each coordinate in the line."""

      return tangents(self.coordinates)

   def splined(self,
               tangent_weight = 'square',
               min_subdivisions = 1,
               max_segment_length = None,
               max_degrees_per_knot = 5.0,
               title = None,
               rep_int_root = None):
      """Retrurns a new Polyline being a cubic spline of this polyline.

        :meta common:
        """

      spline_coords = spline(self.coordinates,
                             tangent_weight = tangent_weight,
                             min_subdivisions = min_subdivisions,
                             max_segment_length = max_segment_length,
                             max_degrees_per_knot = max_degrees_per_knot,
                             closed = self.isclosed)

      if not title:
         title = self.title
      if rep_int_root is None:
         rep_int_root = self.rep_int_root  # todo: check whether it is legal to have 2 representations for 1 interpretation

      return Polyline(self.model,
                      set_bool = self.isclosed,
                      set_coord = spline_coords,
                      set_crs = self.crs_uuid,
                      title = title,
                      rep_int_root = rep_int_root)

   def create_xml(self,
                  ext_uuid = None,
                  add_as_part = True,
                  add_relationships = True,
                  root = None,
                  title = None,
                  originator = None):
      """Create xml from polyline.

        args:
            ext_uuid: the uuid of the hdf5 external part

        :meta common:
        """

      if ext_uuid is None:
         ext_uuid = self.model.h5_uuid()

      if title is not None:
         self.title = title
      if self.title is None:
         self.title = 'polyline'

      polyline = super().create_xml(add_as_part = False, originator = originator)

      if self.rep_int_root is not None:
         rep_int = self.rep_int_root
         if "FaultInterpretation" in str(rqet.content_type(rep_int)):
            content_type = 'obj_FaultInterpretation'
         else:
            content_type = 'obj_HorizonInterpretation'
         self.model.create_ref_node('RepresentedInterpretation',
                                    rqet.citation_title_for_node(rep_int),
                                    rep_int.attrib['uuid'],
                                    content_type = content_type,
                                    root = polyline)

      isclosed = rqet.SubElement(polyline, ns['resqml2'] + 'IsClosed')
      isclosed.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
      isclosed.text = str(self.isclosed).lower()

      nodepatch = rqet.SubElement(polyline, ns['resqml2'] + 'NodePatch')
      nodepatch.set(ns['xsi'] + 'type', ns['resqml2'] + 'NodePatch')
      nodepatch.text = '\n'

      patchindex = rqet.SubElement(nodepatch, ns['resqml2'] + 'PatchIndex')
      patchindex.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
      patchindex.text = str(self.nodepatch[0])

      count = rqet.SubElement(nodepatch, ns['resqml2'] + 'Count')
      count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
      count.text = str(self.nodepatch[1])

      geom = rqet.SubElement(nodepatch, ns['resqml2'] + 'Geometry')
      geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
      geom.text = '\n'

      self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

      points = rqet.SubElement(geom, ns['resqml2'] + 'Points')
      points.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
      points.text = '\n'

      coords = rqet.SubElement(points, ns['resqml2'] + 'Coordinates')
      coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
      coords.text = rqet.null_xml_text

      self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points_patch0', root = coords)

      if root is not None:
         root.append(polyline)
      if add_as_part:
         self.model.add_part('obj_PolylineRepresentation', self.uuid, polyline)
         if add_relationships:
            self.model.create_reciprocal_relationship(polyline, 'destinationObject', self.crs_root, 'sourceObject')
            if self.rep_int_root is not None:  # Optional
               self.model.create_reciprocal_relationship(polyline, 'destinationObject', self.rep_int_root,
                                                         'sourceObject')

            ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
            ext_node = self.model.root_for_part(ext_part)
            self.model.create_reciprocal_relationship(polyline, 'mlToExternalPartProxy', ext_node,
                                                      'externalPartProxyToMl')

      return polyline

   def write_hdf5(self, file_name = None, mode = 'a'):
      """Create or append the coordinates hdf5 array to hdf5 file.

        :meta common:
        """

      if self.uuid is None:
         self.uuid = bu.new_uuid()
      h5_reg = rwh5.H5Register(self.model)
      h5_reg.register_dataset(self.uuid, 'points_patch0', self.coordinates)
      h5_reg.write(file_name, mode = mode)

   def append_extra_metadata(self, meta_dict):
      """Append a given dictionary of metadata to the existing metadata."""

      if self.extra_metadata is None:
         self.extra_metadata = {}
      for key in meta_dict:
         self.extra_metadata[key] = meta_dict[key]


class PolylineSet(_BasePolyline):
   """Class for RESQML polyline set representation."""

   resqml_type = 'PolylineSetRepresentation'

   def __init__(self,
                parent_model,
                set_root = None,
                uuid = None,
                polylines = None,
                irap_file = None,
                charisma_file = None,
                title = None,
                originator = None,
                extra_metadata = None):
      """Initialises a new PolylineSet object.

        arguments:
            parent_model (model.Model object): the model which the new PolylineSetRepresentation belongs to
            set_root (DEPRECATED): use uuid instead;
                the root node of the xml tree representing the PolylineSetRepresentation;
                if not None, the new PolylineSetRepresentation object is initialised based on data in tree;
                if None, expectes a list of polyline objects
            uuid (uuid.UUID, optional): the uuid of an existing RESQML PolylineSetRepresentation object from
                which to initialise this resqpy PolylineSet
            polylines (optional): list of polyline objects from which to build the polylineset
            irap_file (str, optional): the name of a file in irap format from which to import the polyline set
            charisma_file (str, optional): the name of a file in charisma format from which to import the polyline set
            title (str, optional): the citation title to use for a new polyline set;
                ignored if uuid or set_root is not None
            originator (str, optional): the name of the person creating the polyline set, defaults to login id;
                ignored if uuid or set_root is not None
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the polyline set;
                ignored if uuid or set_root is not None

        returns:
            the newly instantiated PolylineSet object

        :meta common:
        """

      self.model = parent_model
      self.coordinates = None
      self.count_perpol = None
      self.polys = []
      self.rep_int_root = None
      self.save_polys = False
      self.boolnotconstant = None
      self.boolvalue = None
      self.crs_uuid = None

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       originator = originator,
                       root_node = set_root,
                       extra_metadata = extra_metadata)

      if self.root is not None:
         return

      if polylines is not None:  # Create from list of polylines
         crs_list = []
         for poly in polylines:
            crs_list.append(poly.crs_uuid)
         crs_set = set(crs_list)
         assert len(crs_set) == 1, 'More than one CRS found in input polylines for polyline set'
         for crs_uuid in crs_set:
            self.crs_uuid = crs_uuid
            if self.crs_root is not None:
               break
         self.polys = polylines
         # Setting the title of the first polyline given as the PolylineSet title
         if len(polylines) > 1:
            self.title = f"{polylines[0].title} + {len(polylines)-1} polylines"
         else:
            self.title = polylines[0].title

      elif irap_file is not None:  # Create from an input IRAP file
         inpoints = rsl.read_lines(irap_file)
         self.count_perpol = []
         closed_array = []
         self.title = os.path.basename(irap_file).split(".")[0]
         for i, poly in enumerate(inpoints):
            if len(poly) > 1:  # Polylines must have at least 2 points
               self.count_perpol.append(len(poly))
               if vu.isclose(poly[0], poly[-1]):
                  closed_array.append(True)
               else:
                  closed_array.append(False)
               if i == 0:
                  self.coordinates = poly
               else:
                  self.coordinates = np.concatenate((self.coordinates, poly))
         self.count_perpol = np.array(self.count_perpol)
         if self.crs_root is None:  # If no crs_uuid is provided, assume the main model crs is valid
            self.crs_uuid = self.model.crs_uuid
         self.polys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                                self.crs_root, self.rep_int_root)

      elif charisma_file is not None:
         with open(charisma_file) as f:
            inpoints = f.readlines()
         self.count_perpol = []
         closed_array = []
         self.title = os.path.basename(charisma_file).split(".")[0]
         for i, line in enumerate(inpoints):
            line = line.split()
            if i == 0:
               self.coordinates = (np.array([[float(line[3]), float(line[4]), float(line[5])]]))
               stick = line[7]
               count = 1
            else:
               self.coordinates = np.concatenate(
                  (self.coordinates, np.array(([[float(line[3]), float(line[4]),
                                                 float(line[5])]]))))
               count += 1
               if stick != line[7] or i == len(inpoints) - 1:
                  if count <= 2:  # Line has fewer than 2 points
                     log.info(f"Polylines must contain at least 2 points - ignoring point {self.coordinates[-2]}")
                     self.coordinates = np.delete(self.coordinates, -2, 0)  # Remove the second to last entry
                  else:
                     self.count_perpol.append(count - 1)
                     closed_array.append(False)
                  count = 1
                  stick = line[7]
         self.count_perpol = np.array(self.count_perpol)
         if self.crs_root is None:  # If no crs_uuid is provided, assume the main model crs is valid
            self.crs_uuid = self.model.crs_uuid
         self.polys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                                self.crs_root, self.rep_int_root)

   def _load_from_xml(self):

      assert self.root is not None  # polyline set xml node specified
      root = self.root

      self.rep_int_root = self.model.referenced_node(rqet.find_tag(root, 'RepresentedInterpretation'))

      for patch_node in rqet.list_of_tag(root, 'LinePatch'):  # Loop over all LinePatches - likely just the one
         assert patch_node is not None  # Required field

         geometry_node = rqet.find_tag(patch_node, 'Geometry')
         assert geometry_node is not None  # Required field

         crs_root = self.model.referenced_node(rqet.find_tag(geometry_node, 'LocalCrs'))
         assert crs_root is not None  # Required field
         self.crs_uuid = rqet.uuid_for_part_root(crs_root)
         assert self.crs_uuid is not None  # Required field

         closed_node = rqet.find_tag(patch_node, 'ClosedPolylines')
         assert closed_node is not None  # Required field
         # The ClosedPolylines could be a BooleanConstantArray, or a BooleanArrayFromIndexArray
         closed_array = self.get_bool_array(closed_node)

         count_node = rqet.find_tag(patch_node, 'NodeCountPerPolyline')
         load_hdf5_array(self, count_node, 'count_perpol', tag = 'Values')

         points_node = rqet.find_tag(geometry_node, 'Points')
         assert points_node is not None  # Required field
         load_hdf5_array(self, points_node, 'coordinates', tag = 'Coordinates')

         # Check that the number of bools aligns with the number of count_perpoly
         # Check that the total of the count_perpoly aligns with the number of coordinates
         assert len(self.count_perpol) == len(closed_array)
         assert np.sum(self.count_perpol) == len(self.coordinates)

         subpolys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                              self.crs_root, self.rep_int_root)
         # Check we have the right number of polygons
         assert len(subpolys) == len(self.count_perpol)

         # Remove duplicate coordinates and count arrays (exist in polylines now)
         # delattr(self,'coordinates')
         # delattr(self,'count_perpol')

         self.polys.extend(subpolys)

   @property
   def crs_root(self):
      """XML node corresponding to self.crs_uuid"""

      return self.model.root_for_uuid(self.crs_uuid)

   def poly_index_containing_point_in_xy(self, p, mode = 'crossing'):
      """Returns the index of the first (closed) polyline containing point p in the xy plane, or None.

        :meta common:
        """

      assert mode in ['crossing', 'winding'], 'unrecognised mode when looking for polygon containing point'

      for i, poly in enumerate(self.polys):
         if not poly.isclosed:
            continue
         if poly.point_is_inside_xy(p, mode = mode):
            return i
      return None

   def create_xml(self,
                  ext_uuid = None,
                  add_as_part = True,
                  add_relationships = True,
                  root = None,
                  title = None,
                  originator = None,
                  save_polylines = False):
      """Create xml from polylineset

        args:
            save_polylines: If true, polylines are also saved individually

        :meta common:
        """

      if ext_uuid is None:
         ext_uuid = self.model.h5_uuid()

      self.save_polys = save_polylines

      if title:
         self.title = title
      if not self.title:
         self.title = 'polyline set'

      if self.save_polys:
         for poly in self.polys:
            poly.create_xml(ext_uuid, add_relationships = add_relationships, originator = originator)

      polyset = super().create_xml(add_as_part = False, originator = originator)

      if self.rep_int_root is not None:
         rep_int = self.rep_int_root
         if "FaultInterpretation" in str(rqet.content_type(rep_int)):
            content_type = 'obj_FaultInterpretation'
         else:
            content_type = 'obj_HorizonInterpretation'
         self.model.create_ref_node('RepresentedInterpretation',
                                    rqet.citation_title_for_node(rep_int),
                                    rep_int.attrib['uuid'],
                                    content_type = content_type,
                                    root = polyset)

      # We convert all Polylines to the CRS of the first Polyline in the set, so set this as crs_uuid

      patch = rqet.SubElement(polyset, ns['resqml2'] + 'LinePatch')
      patch.set(ns['xsi'] + 'type', ns['resqml2'] + 'PolylineSetPatch')
      patch.text = '\n'

      pindex = rqet.SubElement(patch, ns['resqml2'] + 'PatchIndex')
      pindex.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
      pindex.text = '0'

      if self.boolnotconstant:
         # We have mixed data - use a BooleanArrayFromIndexArray
         closed = rqet.SubElement(patch, ns['resqml2'] + 'ClosedPolylines')
         closed.set(ns['xsi'] + 'type', ns['xsd'] + 'BooleanArrayFromIndexArray')
         closed.text = '\n'

         bool_val = rqet.SubElement(closed, ns['resqml2'] + 'Value')
         bool_val.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
         bool_val.text = str(self.boolvalue).lower()

         ind_val = rqet.SubElement(closed, ns['resqml2'] + 'Indices')
         ind_val.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
         ind_val.text = '\n'

         count = rqet.SubElement(closed, ns['resqml2'] + 'Count')
         count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
         count.text = str(len(self.count_perpol))

         self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'indices_patch0', root = ind_val)
      else:
         # All bools are the same - use a BooleanConstantArray
         closed = rqet.SubElement(patch, ns['resqml2'] + 'ClosedPolylines')
         closed.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanConstantArray')
         closed.text = '\n'
         bool_val = rqet.SubElement(closed, ns['resqml2'] + 'Value')
         bool_val.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
         bool_val.text = str(self.boolvalue).lower()
         count = rqet.SubElement(closed, ns['resqml2'] + 'Count')
         count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
         count.text = str(len(self.count_perpol))

      count_pp = rqet.SubElement(patch, ns['resqml2'] + 'NodeCountPerPolyline')
      count_pp.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
      count_pp.text = '\n'

      null = rqet.SubElement(count_pp, ns['resqml2'] + 'NullValue')
      null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
      null.text = '0'

      count_val = rqet.SubElement(count_pp, ns['resqml2'] + 'Values')
      count_val.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
      count_val.text = '\n'

      self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'NodeCountPerPolyline_patch0', root = count_val)

      geom = rqet.SubElement(patch, ns['resqml2'] + 'Geometry')
      geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
      geom.text = '\n'

      self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

      points = rqet.SubElement(geom, ns['resqml2'] + 'Points')
      points.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
      points.text = '\n'

      coords = rqet.SubElement(points, ns['resqml2'] + 'Coordinates')
      coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
      coords.text = '\n'

      self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points_patch0', root = coords)

      if root is not None:
         root.append(polyset)
      if add_as_part:
         self.model.add_part('obj_PolylineSetRepresentation', self.uuid, polyset)
         if add_relationships:
            self.model.create_reciprocal_relationship(polyset, 'destinationObject', self.crs_root, 'sourceObject')
            if self.rep_int_root is not None:  # Optional
               self.model.create_reciprocal_relationship(polyset, 'destinationObject', self.rep_int_root,
                                                         'sourceObject')
            if self.save_polys:
               for poly in self.polys:
                  self.model.create_reciprocal_relationship(polyset, 'destinationObject', poly.root_node,
                                                            'sourceObject')

            ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
            ext_node = self.model.root_for_part(ext_part)
            self.model.create_reciprocal_relationship(polyset, 'mlToExternalPartProxy', ext_node,
                                                      'externalPartProxyToMl')

      return polyset

   def write_hdf5(self, file_name = None, mode = 'a', save_polylines = False):
      """Create or append the coordinates, counts and indices hdf5 arrays to hdf5 file.

        :meta common:
        """

      if self.uuid is None:
         self.uuid = bu.new_uuid()
      self.combine_polylines(self.polys)
      self.bool_array_format(self.closed_array)
      self.save_polys = save_polylines
      if self.save_polys:
         for poly in self.polys:
            poly.write_hdf5(file_name)

      h5_reg = rwh5.H5Register(self.model)
      h5_reg.register_dataset(self.uuid, 'points_patch0', self.coordinates)
      h5_reg.register_dataset(self.uuid, 'NodeCountPerPolyline_patch0', self.count_perpol.astype(np.int32))
      if self.boolnotconstant:
         h5_reg.register_dataset(self.uuid, 'indices_patch0', self.indices)
      h5_reg.write(file_name, mode = mode)

   def get_bool_array(self, closed_node):
      # TODO: Check if also defined boolean arrays
      """ Returns a boolean array using details in the node location.

        If type of boolean array is BooleanConstantArray, uses the array value and count to generate the array. If type of boolean array is BooleanArrayFromIndexArray, find the "other" value bool and indices of the "other" values, and insert these into an array opposite to the main bool.

        args:
            closed_node: the node under which the boolean array information sits"""
      if rqet.node_type(closed_node) == 'BooleanConstantArray':
         count = rqet.find_tag_int(closed_node, 'Count')
         value = rqet.bool_from_text(rqet.node_text(rqet.find_tag(closed_node, 'Value')))
         return np.full((count), value)
      elif rqet.node_type(closed_node) == 'BooleanArrayFromIndexArray':
         count = rqet.find_tag_int(closed_node, 'Count')
         indices_arr = load_hdf5_array(self, closed_node, 'indices_arr', tag = 'Indices')
         istrue = rqet.bool_from_text(rqet.node_text(rqet.find_tag(closed_node, 'IndexIsTrue')))
         out = np.full((count), not istrue)
         out[indices_arr] = istrue
         return out

   def convert_to_polylines(
         self,
         closed_array = None,
         count_perpol = None,
         coordinates = None,
         crs_uuid = None,
         crs_root = None,  # deprecated
         rep_int_root = None):
      """Returns a list of Polylines objects from a PolylineSet

        note:
            all arguments are optional and by default the data will be taken from self

        args:
            closed_array: array containing a bool for each polygon in if it is open (False) or closed (True)
            count_perpol: array containing a list of polygon "lengths" for each polygon
            coordinates: array containing coordinates for all the polygons
            crs_uuid: crs_uuid for polylineset
            crs_root: DEPRECATED; ignored
            rep_int_root: represented interpretation root (optional)

        returns:
            list of polyline objects

        :meta common:
        """

      if count_perpol is None:
         count_perpol = self.count_perpol
      if closed_array is None:
         closed_array = np.zeros(count_perpol, dtype = bool)
         closed_node = rqet.find_nested_tag(self.root, ['LinePatch', 'ClosedPolylines'])
         if closed_node is not None:
            closed_array[:] = self.get_bool_array(closed_node)
      if coordinates is None:
         coordinates = self.coordinates
      if crs_uuid is None:
         crs_uuid = self.crs_uuid
      if rep_int_root is None:
         rep_int_root = self.rep_int_root
      polys = []
      count = 0
      for i in range(len(count_perpol)):
         if i != len(count_perpol) - 1:
            subset = coordinates[count:int(count_perpol[i]) + count].copy()
         else:
            subset = coordinates[count:int(count_perpol[i]) + count + 1].copy()
         if vu.isclose(subset[0], subset[-1]):
            isclosed = True
         else:
            isclosed = closed_array[i]
         count += int(count_perpol[i])
         subtitle = f"{self.title} {i+1}"
         polys.append(
            Polyline(self.model,
                     poly_root = None,
                     set_bool = isclosed,
                     set_coord = subset,
                     set_crs = crs_uuid,
                     title = subtitle,
                     rep_int_root = rep_int_root))

      return polys

   def combine_polylines(self, polylines):
      """Combines the isclosed boolean array, coordinates and count data for a list of polyline objects

        args:
            polylines: list of polyline objects
        """

      self.count_perpol = []
      self.closed_array = []

      for poly in polylines:
         if poly == polylines[0]:
            master_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
            self.crs_uuid = poly.crs_uuid
            self.coordinates = poly.coordinates.copy()
         else:
            curr_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
            if not curr_crs.is_equivalent(master_crs):
               shifted = curr_crs.convert_array_to(master_crs, poly.coordinates)
               self.coordinates = np.concatenate((self.coordinates, shifted))
            else:
               self.coordinates = np.concatenate((self.coordinates, poly.coordinates))

         self.closed_array.append(poly.isclosed)
         self.count_perpol.append(int(len(poly.coordinates)))

      self.count_perpol = np.array(self.count_perpol)

      assert len(self.closed_array) == len(self.count_perpol)
      assert np.sum(self.count_perpol) == len(self.coordinates)

   def bool_array_format(self, closed_array):
      """Determines an appropriate output boolean array format from an input array of bools

        self.boolnotconstant - set to True if all are not open or all closed
        self.boolvalue - value of isclosed for all polylines, or for the majority of polylines if mixed
        self.indices - array of indices where the values are not self.boolvalue, if the polylines are mixed
        """

      self.indices = []
      self.boolnotconstant = False
      if all(closed_array):
         self.boolvalue = True
      elif not all(closed_array) and not any(closed_array):
         self.boolvalue = False
      else:
         if np.count_nonzero(closed_array) > (len(closed_array) / 2):
            self.boolvalue = True
            for i, val in enumerate(closed_array):
               if not val:
                  self.indices.append(i)
         else:
            self.boolvalue = False
            for i, val in enumerate(closed_array):
               if val:
                  self.indices.append(i)
      if len(self.indices) > 0:
         self.boolnotconstant = True

   def set_interpretation_root(self, rep_int_root, recursive = True):
      """Updates the rep_int_root for the polylineset

        args:
            rep_int_root: new rep_int_root
            recursive: boolean, if true will update individual polys with same root
        """

      self.rep_int_root = rep_int_root

      if recursive:
         for poly in self.polys:
            poly.rep_int_root = rep_int_root

   def convert_to_irap(self, file_name):
      """Output an irap file from a polyline set.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name for polyline set representation
        """

      end_of_line = np.array([[999.0, 999.0, 999.0]])
      for poly in self.polys:
         if poly == self.polys[0]:
            out_coords = poly.coordinates
         else:
            out_coords = np.concatenate((out_coords, poly.coordinates))
         out_coords = np.concatenate((out_coords, end_of_line))
      np.savetxt(file_name, out_coords, delimiter = ' ')

   def convert_to_charisma(self, file_name):
      """Output to Charisma fault sticks from a polyline set.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name for polyline set representation
        """

      faultname = self.title.replace(" ", "_")
      lines = []
      for i, poly in enumerate(self.polys):
         for point in poly.coordinates:
            lines.append(f"INLINE-\t0\t0\t{point[0]}\t{point[1]}\t{point[2]}\t{faultname}\t{i+1}\n")
      with open(file_name, 'w') as f:
         for item in lines:
            f.write(item)

   def append_extra_metadata(self, meta_dict):
      """Append a given dictionary of metadata to the existing metadata."""

      if self.extra_metadata is None:
         self.extra_metadata = {}
      for key in meta_dict:
         self.extra_metadata[key] = meta_dict[key]


def shift_polyline(parent_model, poly_root, xyz_shift = (0, 0, 0), title = ''):
   """Returns a new polyline object, shifted by given coordinates."""

   poly = Polyline(parent_model = parent_model, poly_root = poly_root)
   if title != '':
      poly.title = title
   else:
      poly.title = poly.title + f" shifted by xyz({xyz_shift})"
   poly.uuid = bu.new_uuid()
   poly.coordinates = np.array(xyz_shift) + poly.coordinates
   return poly


def flatten_polyline(parent_model, poly_root, axis = "z", value = 0.0, title = ''):
   """Returns a new polyline object, flattened (projected) on a chosen axis to a given value."""

   axis = axis.lower()
   value = float(value)
   assert axis in ["x", "y", "z"], 'Axis must be x, y or z'
   poly = Polyline(parent_model = parent_model, poly_root = poly_root)
   if title != '':
      poly.title = title
   else:
      poly.title = poly.title + f" flattened in {axis} to value {value:.3f}"
   poly.uuid = bu.new_uuid()
   index = "xyz".index(axis)
   poly.coordinates[..., index] = value
   return poly


def tangents(points, weight = 'linear', closed = False):
   """Returns a numpy array of tangent unit vectors for an ordered list of points.

    arguments:
        points (numpy float array of shape (N, 3)): points defining a line
        weight (string, default 'linear'): one of 'linear', 'square' or 'cube', giving
            increased weight to relatively shorter of 2 line segments at each knot
        closed (boolean, default False): if True, the points are treated as a closed
            polyline with regard to end point tangents, otherwise as an open line

    returns:
        numpy float array of the same shape as points, containing a unit length tangent
        vector for each knot (point)

    note:
        if two neighbouring points are identical, a divide by zero will occur
    """

   def one_tangent(points, k1, k2, k3, weight):
      v1 = points[k2] - points[k1]
      v2 = points[k3] - points[k2]
      l1 = vu.naive_length(v1)
      l2 = vu.naive_length(v2)
      if weight == 'square':
         return vu.unit_vector(v1 / (l1 * l1) + v2 / (l2 * l2))
      elif weight == 'cube':
         return vu.unit_vector(v1 / (l1 * l1 * l1) + v2 / (l2 * l2 * l2))
      else:  # linear weight mode
         return vu.unit_vector(v1 / l1 + v2 / l2)

   assert points.ndim == 2 and points.shape[1] == 3
   assert weight in ['linear', 'square', 'cube']

   knot_count = len(points)
   assert knot_count > 1
   tangent_vectors = np.empty((knot_count, 3))

   for knot in range(1, knot_count - 1):
      tangent_vectors[knot] = one_tangent(points, knot - 1, knot, knot + 1, weight)
   if closed:
      assert knot_count > 2, 'closed poly line must contain at least 3 knots for tangent generation'
      tangent_vectors[0] = one_tangent(points, -1, 0, 1, weight)
      tangent_vectors[-1] = one_tangent(points, -2, -1, 0, weight)
   else:
      tangent_vectors[0] = vu.unit_vector(points[1] - points[0])
      tangent_vectors[-1] = vu.unit_vector(points[-1] - points[-2])

   return tangent_vectors


def spline(points,
           tangent_vectors = None,
           tangent_weight = 'square',
           min_subdivisions = 1,
           max_segment_length = None,
           max_degrees_per_knot = 5.0,
           closed = False):
   """Returns a numpy array containing resampled cubic spline of line defined by points.

    arguments:
        points (numpy float array of shape (N, 3)): points defining a line
        tangent_vectors (numpy float array of shape (N, 3), optional) if present, tangent
            vectors to use in the construction of the cubic spline; if None, tangents are
            calculated
        tangent_weight (string, default 'linear'): one of 'linear', 'square' or 'cube', giving
            increased weight to relatively shorter of 2 line segments at each knot when computing
            tangent vectors; ignored if tangent_vectors is not None
        min_subdivisions (int, default 1): the resulting line will have at least this number
            of segments per original line segment
        max_segment_length (float, optional): if present, resulting line segments will not
            exceed this length by much (see notes)
        max_degrees_per_knot (float, default 5.0): the change in direction at each resulting
            knot will not usually exceed this value (see notes)
        closed (boolean, default False): if True, the points are treated as a closed
            polyline with regard to end point tangents, otherwise as an open line

    returns:
        numpy float array of shape (>=N, 3) being knots on a cubic spline defined by points;
        original points are a subset of returned knots

    notes:
        the max_segment_length argument, if present, is compared with the length of each original
        segment to give a lower bound on the number of derived line segments; as the splined line
        may have extra curvature, the length of individual segments in the returned line can exceed
        the argument value, though usually not by much
        similarly, the max_degrees_per_knot is compared to the original deviations to provide another
        lower bound on the number of derived line segments for each original segment; as the spline
        may divide the segment unequally and also sometimes add loops, the resulting deviations can
        exceed the argument value
    """

   assert points.ndim == 2 and points.shape[1] == 3
   knot_count = len(points)
   assert knot_count > (2 if closed else 1)
   if tangent_vectors is None:
      assert tangent_weight in ['linear', 'square', 'cube']
      tangent_vectors = tangents(points, weight = tangent_weight, closed = closed)
   assert tangent_vectors.shape == points.shape
   assert min_subdivisions >= 1
   assert max_segment_length is None or max_segment_length > 0.0
   assert max_degrees_per_knot is None or max_degrees_per_knot > 0.0

   seg_count = knot_count if closed else knot_count - 1
   seg_lengths = np.empty(seg_count)
   for seg in range(knot_count - 1):
      seg_lengths[seg] = vu.naive_length(points[seg + 1] - points[seg])
   if closed:
      seg_lengths[-1] = vu.naive_length(points[0] - points[-1])

   knot_insertions = np.full(seg_count, min_subdivisions - 1, dtype = int)
   if max_segment_length is not None:
      for seg in range(seg_count):
         sub_count = maths.ceil(seg_lengths[seg] / max_segment_length)
         if sub_count > knot_insertions[seg] + 1:
            knot_insertions[seg] = sub_count - 1
   if max_degrees_per_knot is not None:
      for seg in range(seg_count):
         next = seg + 1 if seg < knot_count - 1 else 0
         segment = points[next] - points[seg]
         turn = (vu.degrees_difference(tangent_vectors[seg], segment) +
                 vu.degrees_difference(segment, tangent_vectors[next]))
         log.debug(f'spline segment {seg}; turn: {turn:4.2f}')
         sub_count = maths.ceil(turn / max_degrees_per_knot)
         if sub_count > knot_insertions[seg] + 1:
            knot_insertions[seg] = sub_count - 1
   insertion_count = np.sum(knot_insertions)
   log.debug(f'{insertion_count} knot insertions for spline')
   spline_knot_count = knot_count + insertion_count

   spline_points = np.empty((spline_knot_count, 3))
   sk = 0
   for knot in range(seg_count):
      next = knot + 1 if knot < knot_count - 1 else 0
      spline_points[sk] = points[knot]
      sk += 1
      for i in range(knot_insertions[knot]):
         t = float(i + 1) / float(knot_insertions[knot] + 1)
         t2 = t * t
         t3 = t * t2
         p = np.zeros(3)
         p += (2.0 * t3 - 3.0 * t2 + 1.0) * points[knot]
         p += (t3 - 2.0 * t2 + t) * tangent_vectors[knot] * seg_lengths[knot]
         p += (-2.0 * t3 + 3.0 * t2) * points[next]
         p += (t3 - t2) * tangent_vectors[next] * seg_lengths[knot]
         spline_points[sk] = p
         sk += 1
   if not closed:
      spline_points[sk] = points[-1]
      sk += 1
   assert sk == spline_knot_count

   return spline_points



import numpy as np
from numpy.testing import assert_array_almost_equal

import resqpy.model as rq
import resqpy.surface
import resqpy.organize
import resqpy.grid
import resqpy.lines as rql
import resqpy.grid_surface as rqgs
import resqpy.olio.uuid as bu


def test_surface(tmp_model):

   # Set up a Surface
   title = 'Mountbatten'
   model = tmp_model
   surf = resqpy.surface.Surface(parent_model = model, title = title)
   surf.create_xml()

   # Add a interpretation
   assert surf.represented_interpretation_root is None
   surf.create_interpretation_and_feature(kind = 'fault')
   assert surf.represented_interpretation_root is not None

   # Check fault can be loaded in again
   model.store_epc()
   fault_interp = resqpy.organize.FaultInterpretation(model, uuid = surf.represented_interpretation_uuid)
   fault_feature = resqpy.organize.TectonicBoundaryFeature(model, uuid = fault_interp.tectonic_boundary_feature.uuid)

   # Check title matches expected title
   assert fault_feature.feature_name == title


def test_faces_for_surface(tmp_model):
   crs = resqpy.crs.Crs(tmp_model)
   crs.create_xml()
   grid = resqpy.grid.RegularGrid(tmp_model, extent_kji = (3, 3, 3), crs_uuid = crs.uuid, set_points_cached = True)
   grid.write_hdf5()
   grid.create_xml(write_geometry = True)
   # todo: create sloping planar surface
   # call find faces for each of 3 different methods
   points = np.zeros((2, 2, 3))
   points[1, :, 1] = 3.0
   points[:, 1, 0] = 3.0
   points[:, 1, 2] = 3.0
   points[:, :, 2] += 0.25
   triangles = np.zeros((2, 3), dtype = int)
   triangles[0] = (0, 1, 2)
   triangles[1] = (3, 1, 2)
   surf = resqpy.surface.Surface(tmp_model, crs_uuid = crs.uuid)
   surf.set_from_triangles_and_points(triangles, points.reshape((-1, 3)))
   assert surf is not None
   gcs = rqgs.find_faces_to_represent_surface(grid, surf, 'staffa', mode = 'staffa')
   assert gcs is not None
   assert gcs.count == 12
   cip = set([tuple(pair) for pair in gcs.cell_index_pairs])
   expected_cip = grid.natural_cell_indices(
      np.array([[[0, 0, 0], [1, 0, 0]], [[0, 1, 0], [1, 1, 0]], [[0, 2, 0], [1, 2, 0]], [[1, 0, 0], [1, 0, 1]],
                [[1, 1, 0], [1, 1, 1]], [[1, 2, 0], [1, 2, 1]], [[1, 0, 1], [2, 0, 1]], [[1, 1, 1], [2, 1, 1]],
                [[1, 2, 1], [2, 2, 1]], [[2, 0, 1], [2, 0, 2]], [[2, 1, 1], [2, 1, 2]], [[2, 2, 1], [2, 2, 2]]],
               dtype = int))
   e_cip = set([tuple(pair) for pair in expected_cip])
   assert cip == e_cip  # note: this assumes lower cell index is first, which happens to be true
   # todo: check face indices
   gcs.write_hdf5()
   gcs.create_xml()
   assert bu.matching_uuids(tmp_model.uuid(obj_type = 'GridConnectionSetRepresentation'), gcs.uuid)


def test_delaunay_triangulation(example_model_and_crs):

   model, crs = example_model_and_crs

   # number of random points to use
   n = 20

   # create a set of random points
   x = np.random.random(n) * 1000.0
   y = np.random.random(n) * 1000.0
   z = np.random.random(n)  # note: triangulation does not use z values
   p = np.stack((x, y, z), axis = -1)

   # make a PointSet object
   ps = resqpy.surface.PointSet(model, crs_uuid = crs.uuid, points_array = p, title = 'random points in square')

   # make another PointSet as random points within a closed polyline
   vertices = np.array(
      ((50.0, 99.0, 13.0), (85.0, 60.0, 17.5), (62.7, 11.0, 10.0), (33.3, 15.3, 19.2), (12.8, 57.8, 15.0)))
   polygon = rql.Polyline(model, set_crs = crs.uuid, set_bool = True, set_coord = vertices, title = 'the pentagon')
   polygon.write_hdf5()
   polygon.create_xml()
   ps2 = resqpy.surface.PointSet(model,
                                 crs_uuid = crs.uuid,
                                 polyline = polygon,
                                 random_point_count = n,
                                 title = 'random points in polygon')

   # process the point sets into triangulated surfaces
   for point_set in (ps, ps2):
      point_set.write_hdf5()
      point_set.create_xml()
      surf = resqpy.surface.Surface(model, point_set = point_set, title = 'surface from ' + str(point_set.title))
      assert surf is not None
      surf.write_hdf5()
      surf.create_xml()
      # check that coordinate range of points looks okay
      triangles, points = surf.triangles_and_points()
      assert len(points) == n
      original_points = point_set.full_array_ref()
      assert_array_almost_equal(np.nanmin(original_points, axis = 0), np.nanmin(points, axis = 0))
      assert_array_almost_equal(np.nanmax(original_points, axis = 0), np.nanmax(points, axis = 0))


def test_regular_mesh(example_model_and_crs):

   model, crs = example_model_and_crs

   # number of points in mesh, origin spacing
   ni = 7
   nj = 5
   origin = (409000.0, 1605000.0, 0.0)
   di = dj = 50.0

   # create some random depths
   z = (np.random.random(ni * nj) * 20.0 + 1000.0).reshape((nj, ni))

   # make a regular mesh representation
   mesh = resqpy.surface.Mesh(model,
                              crs_uuid = crs.uuid,
                              mesh_flavour = 'reg&z',
                              ni = ni,
                              nj = nj,
                              origin = origin,
                              dxyz_dij = np.array([[di, 0.0, 0.0], [0.0, dj, 0.0]]),
                              z_values = z,
                              title = 'random mesh',
                              originator = 'Andy',
                              extra_metadata = {'testing mode': 'automated'})
   assert mesh is not None
   mesh.write_hdf5()
   mesh.create_xml()
   mesh_uuid = mesh.uuid

   # fully write model to disc
   model.store_epc()
   epc = model.epc_file

   # re-open model and check the mesh object is there
   model = rq.Model(epc)
   assert bu.matching_uuids(model.uuid(obj_type = 'Grid2dRepresentation', title = 'random mesh'), mesh_uuid)

   # establish a resqpy Mesh from the object in the RESQML dataset
   peristent_mesh = resqpy.surface.Mesh(model, uuid = mesh_uuid)

   # check some of the metadata
   assert peristent_mesh.ni == ni and peristent_mesh.nj == nj
   assert peristent_mesh.flavour == 'reg&z'
   assert_array_almost_equal(np.array(peristent_mesh.regular_origin), np.array(origin))
   assert_array_almost_equal(np.array(peristent_mesh.regular_dxyz_dij), np.array([[di, 0.0, 0.0], [0.0, dj, 0.0]]))

   # check a fully expanded version of the points
   assert_array_almost_equal(peristent_mesh.full_array_ref(), mesh.full_array_ref())

   # check that we can build a Surface from the Mesh
   surf = peristent_mesh.surface(quad_triangles = True)
   assert surf is not None

   # do some basic checks that the surface looks consistent with the mesh
   t, p = surf.triangles_and_points()
   assert len(p) == (ni * nj) + ((ni - 1) * (nj - 1))  # quad triangles mode introduces the extra points
   assert len(t) == 4 * (ni - 1) * (nj - 1)
   assert_array_almost_equal(np.min(p, axis = 0), np.min(peristent_mesh.full_array_ref().reshape(-1, 3), axis = 0))
   assert_array_almost_equal(np.max(p, axis = 0), np.max(peristent_mesh.full_array_ref().reshape(-1, 3), axis = 0))



# grid functions module

version = '29th April 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)
log.debug('grid_functions.py version %s', version)

# defs:
# def infill_block_geometry(extent, depth, thickness, x, y,
#    k_increase_direction = 'down', depth_zero_tolerance = 0.01,
#    x_y_zero_tolerance = 0.01,
#    vertical_cell_overlap_tolerance = 0.01,
#    snap_to_top_and_base = True, nudge = True):
# def resequence_nexus_corp(corner_points, eight_mode = False, undo = False):
# def random_cell(corner_points, border = 0.25, max_tries = 20, tolerance = 0.003):
# def determine_corp_ijk_handedness(corner_points, xyz_is_left_handed = True):
# def determine_corp_extent(corner_points, tolerance = 0.003):
# def translate_corp(corner_points, x_shift = None, y_shift = None, min_xy = 0.0, shift_rounding_digits = None):

import math as maths
import random
import numpy as np

import resqpy.olio.vector_utilities as vec
import resqpy.olio.factors as factors

##########################################################################################
# infill_block_geometry():
# scans each logically vertical column of cells,
# assigning depth and thickness values for inactive cells sandwiched between active cells
# extent is a 3 element vector: nk,nj,ni
# depth is a 3D numpy float array of size matching extent
# depth values assumed more positive with increasing depth
# zero values in depth input array indicate inactive cells
# thickness is a 3D numpy float array of size matching extent
# x and y are each a 3D numpy float array of size matching extent
# k_increase_direction is either 'up' or 'down'
# depth_zero_tolerance is maximum value for which depth is considered zero
# vertical_cell_overlap_tolerance is the maximum acceptable overlap of cells on input
# snap_to_top_and_base, when True, causes cells above topmost active and below deepest active
# to be populated with pinched out cells at the top and bottom faces respectively
# nudge causes the depth of cells with greater k to be moved to clean up overlap over pinchouts


def infill_block_geometry(extent,
                          depth,
                          thickness,
                          x,
                          y,
                          k_increase_direction = 'down',
                          depth_zero_tolerance = 0.01,
                          x_y_zero_tolerance = 0.01,
                          vertical_cell_overlap_tolerance = 0.01,
                          snap_to_top_and_base = True,
                          nudge = True):
   """Scans logically vertical columns of cells setting depth (& thickness) of inactive cells."""

   if k_increase_direction == 'down':
      k_dir_sign = 1.0
   elif k_increase_direction == 'up':
      k_dir_sign = -1.0
   else:
      assert (False)

   for j in range(extent[1]):
      for i in range(extent[2]):
         k_top = 0  # NB: 'top' & 'bottom' are misleading if k_increase_direction == 'up'
         while k_top < extent[0] and abs(depth[k_top, j, i]) <= depth_zero_tolerance:
            depth[k_top, j, i] = 0.0  # clean up tiny values
            thickness[k_top, j, i] = 0.0
            if abs(x[k_top, j, i]) <= x_y_zero_tolerance:
               x[k_top, j, i] = 0.0
            if abs(y[k_top, j, i]) <= x_y_zero_tolerance:
               y[k_top, j, i] = 0.0
            k_top += 1  # skip topmost inactive batch
         if k_top >= extent[0]:
            continue  # whole column is inactive
         if snap_to_top_and_base:
            snap_depth = depth[k_top, j, i] - k_dir_sign * thickness[k_top, j, i] / 2.0
            snap_x = x[k_top, j, i]
            snap_y = y[k_top, j, i]
            for k_snap in range(k_top):
               depth[k_snap, j, i] = snap_depth
               x[k_snap, j, i] = snap_x
               y[k_snap, j, i] = snap_y
         while True:
            while k_top < extent[0] and abs(depth[k_top, j, i]) > depth_zero_tolerance:  # skip active layers
               k_top += 1
            k_base = k_top + 1
            while k_base < extent[0] and abs(depth[k_base, j, i]) <= depth_zero_tolerance:
               depth[k_base, j, i] = 0.0  # clean up tiny depth values
               thickness[k_base, j, i] = 0.0
               if abs(x[k_base, j, i]) <= x_y_zero_tolerance:
                  x[k_base, j, i] = 0.0
               if abs(y[k_base, j, i]) <= x_y_zero_tolerance:
                  y[k_base, j, i] = 0.0
               k_base += 1  # look for deeper active layer
            if k_base >= extent[0]:  # no deeper active cells found
               if snap_to_top_and_base:
                  snap_depth = depth[k_top - 1, j, i] + k_dir_sign * thickness[k_top - 1, j, i] / 2.0
                  snap_x = x[k_top - 1, j, i]
                  snap_y = y[k_top - 1, j, i]
                  for k_snap in range(extent[0] - k_top):
                     depth[k_top + k_snap, j, i] = snap_depth
                     x[k_top + k_snap, j, i] = snap_x
                     y[k_top + k_snap, j, i] = snap_y
               break
            void_cell_count = k_base - k_top
            assert (void_cell_count > 0)
            void_top_depth = depth[k_top - 1, j, i] + (thickness[k_top - 1, j, i] / 2.0) * k_dir_sign
            void_bottom_depth = depth[k_base, j, i] - (thickness[k_base, j, i] / 2.0) * k_dir_sign
            void_top_x = x[k_top - 1, j, i]
            void_top_y = y[k_top - 1, j, i]
            void_interval = k_dir_sign * (void_bottom_depth - void_top_depth)
            void_x_interval = x[k_base, j, i] - void_top_x
            void_y_interval = y[k_base, j, i] - void_top_y
            infill_cell_thickness = void_interval / void_cell_count
            if void_interval < 0.0:  # overlapping cells
               if -void_interval < vertical_cell_overlap_tolerance:
                  if nudge:
                     nudge_count = 0  # debug
                     for k_nudge in range(extent[0] - k_base):
                        if depth[k_base + k_nudge, j, i] > depth_zero_tolerance:
                           depth[k_base + k_nudge, j, i] += -void_interval * k_dir_sign
                           nudge_count += 1  # debug
                     log.debug('%1d cells nudged in [ i j ] column [%1d, %1d]', nudge_count, i + 1, j + 1)
                     void_bottom_depth += -void_interval
                  void_interval = 0.0
                  infill_cell_thickness = 0.0
               else:
                  log.warn('Cells [%1d, %1d, %1d] and [%1d, %1d, %1d] overlap ...', i + 1, j + 1, k_top, i + 1, j + 1,
                           k_base + 1)
                  log.warn('   check k_increase_direction and tolerances')
                  log.warn('Skipping rest of i,j column')  # todo: could abort here
                  break
            assert (infill_cell_thickness >= 0.0)
            for void_k in range(void_cell_count):
               depth[k_top + void_k, j, i] = void_top_depth + (0.5 + void_k) * infill_cell_thickness * k_dir_sign
               thickness[k_top + void_k, j, i] = infill_cell_thickness
               x[k_top + void_k, j, i] = void_top_x + (0.5 + void_k) * void_x_interval / void_cell_count
               y[k_top + void_k, j, i] = void_top_y + (0.5 + void_k) * void_y_interval / void_cell_count
            k_top = k_base


# end of def infill_block_geometry()
##########################################################################################

##########################################################################################
# def resequence_nexus_corp():


def resequence_nexus_corp(corner_points, eight_mode = False, undo = False):
   """Reorders corner point data in situ, to handle bizarre nexus orderings"""

   # undo False for corp to internal; undo True for internal to corp; only relevant in eight_mode
   assert (corner_points.ndim == 7)
   extent = np.array(corner_points.shape, dtype = 'int')
   if eight_mode:
      for k in range(extent[0]):
         for j in range(extent[1]):
            for i in range(extent[2]):
               if undo:
                  xyz = np.zeros((3, 8))
                  c = 0
                  for kp in range(2):
                     for jp in range(2):
                        for ip in range(2):
                           xyz[:, c] = corner_points[k, j, i, kp, jp, ip, :]
                           c += 1
                  corner_points[k, j, i] = xyz.reshape((2, 2, 2, 3))
               else:
                  xyz = corner_points[k, j, i].reshape((3, 8)).copy()
                  c = 0
                  for kp in range(2):
                     for jp in range(2):
                        for ip in range(2):
                           corner_points[k, j, i, kp, jp, ip, :] = xyz[:, c]
                           c += 1
   else:  # reversible, so not dependent on undo argument
      jp_slice = corner_points[:, :, :, :, 1, 0, :].copy()
      corner_points[:, :, :, :, 1, 0, :] = corner_points[:, :, :, :, 1, 1, :]
      corner_points[:, :, :, :, 1, 1, :] = jp_slice


# end of def resequence_nexus_corp()
##########################################################################################

##########################################################################################
# def random_cell():


def random_cell(corner_points, border = 0.25, max_tries = 20, tolerance = 0.003):
   """Returns a random cell's (k,j,i) tuple for a cell with non-zero lengths on all 3 primary edges."""

   assert (corner_points.ndim == 7)
   assert (border >= 0.0 and border < 0.5)
   assert (max_tries > 0)

   extent = np.array(corner_points.shape, dtype = 'int')
   kji_extent = extent[:3]
   kji_border = np.zeros(3, dtype = 'int')
   kji_upper = np.zeros(3, dtype = 'int')
   for axis in range(3):
      kji_border[axis] = int(float(kji_extent[axis]) * border)
      kji_upper[axis] = kji_extent[axis] - kji_border[axis] - 1
      if kji_upper[axis] < kji_border[axis]:
         kji_upper[axis] = kji_border[axis]

   kji_cell = np.empty(3, dtype = 'int')
   attempt = 0
   while attempt < max_tries:
      attempt += 1
      for axis in range(3):
         if kji_extent[axis] == 1:
            kji_cell[axis] = 0
         else:
            kji_cell[axis] = random.randint(kji_border[axis], kji_upper[axis])
      cell_cp = corner_points[tuple(kji_cell)]
      assert (cell_cp.shape == (2, 2, 2, 3))
      if vec.manhatten_distance(cell_cp[0, 0, 0], cell_cp[1, 0, 0]) < tolerance:
         continue
      if vec.manhatten_distance(cell_cp[0, 0, 0], cell_cp[0, 1, 0]) < tolerance:
         continue
      if vec.manhatten_distance(cell_cp[0, 0, 0], cell_cp[0, 0, 1]) < tolerance:
         continue
      return tuple(kji_cell)

   log.warning('failed to find random voluminous cell')
   return None


# end of def random_cell()
##########################################################################################

##########################################################################################
# def determine_corp_ijk_handedness():


def determine_corp_ijk_handedness(corner_points, xyz_is_left_handed = True):
   """Determine true ijk handedness from corner point data in pagoda style 7D array; returns 'right' or 'left'."""

   assert (corner_points.ndim == 7)
   cell_kji = random_cell(corner_points)
   assert (cell_kji is not None)
   log.debug('using cell ijk0 [{}, {}, {}] to determine ijk handedness'.format(cell_kji[2], cell_kji[1], cell_kji[0]))
   cell_cp = corner_points[cell_kji]
   origin = cell_cp[0, 0, 0]
   det = vec.determinant(cell_cp[0, 0, 1] - origin, cell_cp[0, 1, 0] - origin,
                         cell_cp[1, 0, 0] - origin)  # NB. IJK ordering
   if det == 0.0:
      log.warning('indeterminate handedness in cell ijk0 [{}, {}, {}]'.format(cell_kji[2], cell_kji[1], cell_kji[0]))
      return None
   if det > 0.0:
      ijk_is_left_handed = xyz_is_left_handed
   else:
      ijk_is_left_handed = not xyz_is_left_handed
   if ijk_is_left_handed:
      return 'left'
   return 'right'


# end of def determine_corp_ijk_handedness()
##########################################################################################

##########################################################################################
# def determine_corp_extent():


def determine_corp_extent(corner_points, tolerance = 0.003):
   """Returns extent of grid derived from 7D corner points with all cells temporarily in I."""

   def neighbours(corner_points, sextuple_cell_a_p1, sextuple_cell_a_p2, sextuple_cell_b_p1, sextuple_cell_b_p2,
                  tolerance):
      # allows for reversal of points (or not) in neighbouring cell
      if (
         (vec.manhatten_distance(corner_points[sextuple_cell_a_p1], corner_points[sextuple_cell_b_p1]) <= tolerance) and
         (vec.manhatten_distance(corner_points[sextuple_cell_a_p2], corner_points[sextuple_cell_b_p2]) <= tolerance)):
         return True
      if (
         (vec.manhatten_distance(corner_points[sextuple_cell_a_p1], corner_points[sextuple_cell_b_p2]) <= tolerance) and
         (vec.manhatten_distance(corner_points[sextuple_cell_a_p2], corner_points[sextuple_cell_b_p1]) <= tolerance)):
         return True
      return False

   assert (corner_points.ndim == 7 and corner_points.shape[:2] == (1, 1))

   confirmation = 3  # number of identical results needed for each of NI and NJ
   max_failures = 100  # maximum number of failed random cells for each of NI and NJ
   min_cell_length = 10.0 * tolerance

   cell_count = corner_points.shape[2]
   prime_factorization = factors.factorize(cell_count)
   log.debug('cell count is ' + str(cell_count) + '; prime factorization: ' + str(prime_factorization))
   possible_extents = factors.all_factors_from_primes(prime_factorization)
   log.debug('possible extents are: ' + str(possible_extents))

   ni = None
   redundancy = confirmation
   remaining_attempts = max_failures
   while redundancy:
      kji_cell = random_cell(corner_points, tolerance = min_cell_length)
      found = False
      for e in possible_extents:
         candidate = kji_cell[2] + e
         if candidate >= cell_count:
            continue
         if neighbours(corner_points, (0, 0, kji_cell[2], 0, 1, 0), (0, 0, kji_cell[2], 0, 1, 1),
                       (0, 0, candidate, 0, 0, 0), (0, 0, candidate, 0, 0, 1), tolerance):
            if ni is not None and ni != e:
               log.error('inconsistent NI values of {} and {} determined from corner points'.format(ni, e))
               return None
            found = True
            ni = e
            redundancy -= 1
            break
      if not found:
         remaining_attempts -= 1
         if remaining_attempts <= 0:
            log.error('failed to determine NI from corner points (out of tries)')  # could assume NJ = 1 here
            return None

   log.info('NI determined from corner points to be ' + str(ni))

   if ni > 1:
      ni_prime_factors = factors.factorize(ni)
      factors.remove_subset(prime_factorization, ni_prime_factors)
      log.debug('remaining prime factors after accounting for NI are: ' + str(prime_factorization))
      possible_extents = factors.all_factors_from_primes(prime_factorization)
      log.debug('possible extents for NJ & NK are: ' + str(possible_extents))

   nj = None
   redundancy = confirmation
   remaining_attempts = max_failures
   while redundancy:
      kji_cell = random_cell(corner_points)
      found = False
      for e in possible_extents:
         candidate = kji_cell[2] + (e * ni)
         if candidate >= cell_count:
            continue
         if vec.manhatten_distance(corner_points[0, 0, kji_cell[2], 1, 0, 0], corner_points[0, 0, candidate, 0, 0,
                                                                                            0]) <= tolerance:
            if nj is not None and nj != e:
               log.error('inconsistent NJ values of {} and {} determined from corner points'.format(nj, e))
               return None
            found = True
            nj = e
            redundancy -= 1
            break
      if not found:
         remaining_attempts -= 1
         if remaining_attempts <= 0:
            log.error(
               'failed to determine NJ from corner points (out of tries)')  # could assume or check if NK = 1 here
            return None

   log.info('NJ determined from corner points to be ' + str(nj))

   nk, remainder = divmod(cell_count, ni * nj)
   assert (remainder == 0)
   log.info('NK determined from corner points to be ' + str(nk))
   assert (nk in possible_extents)

   return [nk, nj, ni]


# end def determine_corp_extent():
##########################################################################################

##########################################################################################
# def translate_corp():


def translate_corp(corner_points, x_shift = None, y_shift = None, min_xy = None, preserve_digits = None):
   """Adjusts x and y values of corner points by a constant offset."""

   assert (corner_points.ndim == 7)
   if min_xy is None:
      minimum_xy = 0.0
   else:
      minimum_xy = min_xy
   if x_shift is None:
      x_sub = np.min(corner_points[:, :, :, :, :, :, 0]) - minimum_xy
   else:
      x_sub = -x_shift
   if y_shift is None:
      y_sub = np.min(corner_points[:, :, :, :, :, :, 1]) - minimum_xy
   else:
      y_sub = -y_shift
   if preserve_digits is not None:
      divisor = maths.pow(10.0, preserve_digits)
      x_sub = divisor * maths.floor(x_sub / divisor)
      y_sub = divisor * maths.floor(y_sub / divisor)

   log.info('translating corner points by %3.1f in x and %3.1f in y', -x_sub, -y_sub)
   corner_points[:, :, :, :, :, :, 0] -= x_sub
   corner_points[:, :, :, :, :, :, 1] -= y_sub


# end of def translate_corp()
##########################################################################################


def triangles_for_cell_faces(cp):
   """Returns numpy array of shape (3, 2, 4, 3, 3) with axes being kji, -+, triangle within face, triangle corner, xyz.

   args:
      cp (numpy float array of shape (2, 2, 2, 3)): single cell corner point array in pagoda protocol

   returns:
      numpy float array of shape (3, 2, 4, 3, 3) holding triangle corner coordinates for cell faces represented with
      quad triangles

   note:
      resqpy.surface also contains methods for working with cell faces as triangulated sets
   """

   tri = np.empty((3, 2, 4, 3, 3))

   # create face centre points and assign as one vertex in each of 4 trangles for face
   tri[0, :, :, 0] = np.mean(cp, axis = (1, 2)).reshape((2, 1, 3)).repeat(4, axis = 1).reshape(
      (2, 4, 3))  # k face centres
   tri[1, :, :, 0] = np.mean(cp, axis = (0, 2)).reshape((2, 1, 3)).repeat(4, axis = 1).reshape(
      (2, 4, 3))  # j face centres
   tri[2, :, :, 0] = np.mean(cp, axis = (0, 1)).reshape((2, 1, 3)).repeat(4, axis = 1).reshape(
      (2, 4, 3))  # i face centres

   # k faces
   tri[0, :, 0, 1] = cp[:, 0, 0]
   tri[0, :, 0, 2] = cp[:, 0, 1]
   tri[0, :, 1, 1] = cp[:, 0, 1]
   tri[0, :, 1, 2] = cp[:, 1, 1]
   tri[0, :, 2, 1] = cp[:, 1, 1]
   tri[0, :, 2, 2] = cp[:, 1, 0]
   tri[0, :, 3, 1] = cp[:, 1, 0]
   tri[0, :, 3, 2] = cp[:, 0, 0]

   # j faces
   tri[1, :, 0, 1] = cp[0, :, 0]
   tri[1, :, 0, 2] = cp[0, :, 1]
   tri[1, :, 1, 1] = cp[0, :, 1]
   tri[1, :, 1, 2] = cp[1, :, 1]
   tri[1, :, 2, 1] = cp[1, :, 1]
   tri[1, :, 2, 2] = cp[1, :, 0]
   tri[1, :, 3, 1] = cp[1, :, 0]
   tri[1, :, 3, 2] = cp[0, :, 0]

   # i faces
   tri[2, :, 0, 1] = cp[0, 0, :]
   tri[2, :, 0, 2] = cp[0, 1, :]
   tri[2, :, 1, 1] = cp[0, 1, :]
   tri[2, :, 1, 2] = cp[1, 1, :]
   tri[2, :, 2, 1] = cp[1, 1, :]
   tri[2, :, 2, 2] = cp[1, 0, :]
   tri[2, :, 3, 1] = cp[1, 0, :]
   tri[2, :, 3, 2] = cp[0, 0, :]

   return tri


# end of grid_functions module
##########################################################################################


def actual_pillar_shape(pillar_points, tolerance = 0.001):
   """Returns 'curved', 'straight' or 'vertical' for shape of fully defined points array of shape (nk + k_gaps + 1, ..., 3)."""

   assert pillar_points.ndim >= 3 and pillar_points.shape[-1] == 3

   pp = pillar_points.reshape((pillar_points.shape[0], -1, 3))

   from_top = pp - pp[0]
   xy_drift = np.abs(from_top[:, :, 0]) + np.abs(from_top[:, :,
                                                          1])  # use Manhattan distance as cheap proxy for true distance
   if np.max(xy_drift) <= tolerance:
      return 'vertical'
   if np.max(xy_drift[-1]) <= tolerance:
      return 'curved'  # top & bottom are vertically aligned, so pillar must be curved

   # where z variation is tiny (null pillar), don't interpolate, just treat these pillars as straight
   # elsewhere find drift from vector defined by from_top[-1]
   null_pillar_mask = (abs(from_top[-1, :, 2]) <= tolerance)
   from_top[-1, :, 2] = np.where(null_pillar_mask, tolerance, from_top[-1, :, 2])  # avoid divide by zero issues
   z_fraction = from_top[:, :, 2] / from_top[-1, :, 2]
   xy_drift = from_top[:, :, :2] - z_fraction.reshape((pp.shape[0], pp.shape[1], 1)) * from_top[-1, :, :2].reshape(
      (1, pp.shape[1], 2))
   straight = (np.max(np.sum(np.abs(xy_drift), axis = -1), axis = 0) <= tolerance)
   masked_straight = np.where(null_pillar_mask, True, straight)
   if np.all(masked_straight):
      return 'straight'
   return 'curved'


##########################################################################################


def columns_to_nearest_split_face(grid):
   """Returns a numpy integer array of shape (NJ, NI) being number of cells to nearest split edge (Manhattan distance)."""

   if not grid.has_split_coordinate_lines:
      return None

   j_col_faces_split, i_col_faces_split = grid.split_column_faces()
   abutting = np.zeros((grid.nj, grid.ni), dtype = bool)
   abutting[:-1, :] = j_col_faces_split
   abutting[1:, :] = np.logical_or(abutting[1:, :], j_col_faces_split)
   abutting[:, :-1] = np.logical_or(abutting[:, :-1], i_col_faces_split)
   abutting[:, 1:] = np.logical_or(abutting[:, 1:], i_col_faces_split)
   framed = np.full((grid.nj + 2, grid.ni + 2), grid.nj + grid.ni, dtype = int)
   framed[1:-1, 1:-1] = np.where(abutting, 0, grid.nj + grid.ni)

   while True:
      plus_one = framed + 1
      updated = np.minimum(framed[1:-1, 1:-1], plus_one[:-2, 1:-1])
      updated[:] = np.minimum(updated, plus_one[2:, 1:-1])
      updated[:] = np.minimum(updated, plus_one[1:-1, :-2])
      updated[:] = np.minimum(updated, plus_one[1:-1, 2:])
      if np.all(updated == framed[1:-1, 1:-1]):
         break
      framed[1:-1, 1:-1] = updated

   return framed[1:-1, 1:-1]


##########################################################################################


def left_right_foursome(full_pillar_list, p_index):
   """Returns (2, 2) bool numpy array indicating which columns around a primary pillar are to the right of a line."""

   assert 0 < p_index < len(full_pillar_list) - 1
   here = np.array(full_pillar_list[p_index], dtype = int)
   previous = np.array(full_pillar_list[p_index - 1], dtype = int)
   next = np.array(full_pillar_list[p_index + 1], dtype = int)
   entry = tuple(here - previous)
   exit = tuple(next - here)
   if entry == (0, 1):
      if exit == (-1, 0):
         return np.array([[False, True], [True, True]], dtype = bool)
      elif exit == (0, 1):
         return np.array([[False, False], [True, True]], dtype = bool)
      elif exit == (1, 0):
         return np.array([[False, False], [True, False]], dtype = bool)
      else:
         raise Exception('code failure whilst taking exit sides from dubious full pillar list')
   elif entry == (0, -1):
      if exit == (-1, 0):
         return np.array([[False, True], [False, False]], dtype = bool)
      elif exit == (0, -1):
         return np.array([[True, True], [False, False]], dtype = bool)
      elif exit == (1, 0):
         return np.array([[True, True], [True, False]], dtype = bool)
      else:
         raise Exception('code failure whilst taking exit sides from dubious full pillar list')
   elif entry == (1, 0):
      if exit == (0, -1):
         return np.array([[True, False], [False, False]], dtype = bool)
      elif exit == (1, 0):
         return np.array([[True, False], [True, False]], dtype = bool)
      elif exit == (0, 1):
         return np.array([[True, False], [True, True]], dtype = bool)
      else:
         raise Exception('code failure whilst taking exit sides from dubious full pillar list')
   elif entry == (-1, 0):
      if exit == (0, -1):
         return np.array([[True, True], [False, True]], dtype = bool)
      elif exit == (-1, 0):
         return np.array([[False, True], [False, True]], dtype = bool)
      elif exit == (0, 1):
         return np.array([[False, False], [False, True]], dtype = bool)
      else:
         raise Exception('code failure whilst taking exit sides from dubious full pillar list')
   else:
      log.debug(f'entry pair: {entry}')
      raise Exception('code failure whilst taking entry sides from dubious full pillar list')


##########################################################################################



"""DEPRECATED, use resqpy.weights_and_measures instead"""

import warnings
from resqpy.weights_and_measures import *  # noqa

warnings.warn(
   "resqpy.olio.weights_and_measures is Deprecated and will be removed. Please update your code to use resqpy.weights_and_measures"
)



""" RESQML manipulation library

.. autosummary::
    :toctree: _autosummary
    :caption: API Reference
    :template: custom-module-template.rst
    :recursive:

    model
    crs
    derived_model
    fault
    grid
    grid_surface
    lines
    organize
    property
    rq_import
    surface
    time_series
    weights_and_measures
    well
    olio

"""

try:
   # Version dynamically extracted from git tags when package is built
   from .version import version as __version__  # type: ignore

except ImportError:
   __version__ = "0.0.0-version-not-available"



import resqpy.olio.uuid as bu
from resqpy.olio.base import BaseResqpy


class DummyObj(BaseResqpy):
   resqml_type = 'DummyResqmlInterpretation'


class ReusableDummyObj(BaseResqpy):
   resqml_type = 'ReusableDummyResqmlInterpretation'

   # a class must have this method to allow meaningful reusability
   def is_equivalent(self, other):
      return True


def test_base_creation(tmp_model):

   # Setup new object
   title = 'Wondermuffin'
   dummy = DummyObj(model = tmp_model, title = title)

   # UUID should exist
   # but root should not exist yet and part name does not exist in the parent model
   assert dummy.uuid is not None
   #   assert dummy.part is not None  # this version constructed the part from resqml_type and uuid
   assert dummy.part is None
   assert dummy.root is None

   # After creating XML, root should exist
   dummy.create_xml(add_as_part = True)
   assert dummy.root is not None


def test_base_save_and_load(tmp_model):

   # Create and save a DummyObj
   title = 'feefifofum'
   originator = 'Scruffian'
   metadata = {"balderdash": "pumpernickel"}
   dummy1 = DummyObj(model = tmp_model, title = title, originator = originator)
   dummy1.extra_metadata = metadata

   # Save to XML
   dummy1.create_xml(add_as_part = True)

   # Load a new object
   dummy2 = DummyObj(model = tmp_model, uuid = dummy1.uuid)

   # Properties should match
   assert bu.matching_uuids(dummy2.uuid, dummy1.uuid)
   assert dummy2.root is not None
   assert dummy2.title == title
   assert dummy2.originator == originator
   assert dummy2.extra_metadata == metadata


def test_base_comparison(tmp_model):

   # Setup new object
   dummy1 = DummyObj(model = tmp_model)
   dummy2 = DummyObj(model = tmp_model, uuid = dummy1.uuid)  # Same UUID
   dummy3 = DummyObj(model = tmp_model, title = 'hello')  # Different UUID

   # Comparison should work by UUID
   assert dummy1 == dummy2
   assert dummy1 != dummy3


def test_base_repr(tmp_model):

   dummy = DummyObj(model = tmp_model)

   # Check repr method produces a string
   rep = repr(dummy)
   assert isinstance(rep, str)
   assert len(rep) > 0

   # Check HTML can be generated
   html = dummy._repr_html_()
   assert len(html) > 0


def test_base_reuse_duplicate(tmp_model):

   # Create object and save
   dummy_1 = ReusableDummyObj(model = tmp_model)
   assert not dummy_1.try_reuse()
   dummy_1.create_xml(add_as_part = True)

   # Should be present in model
   uuids = tmp_model.uuids(obj_type = ReusableDummyObj.resqml_type)
   assert len(uuids) == 1
   assert dummy_1.try_reuse()  # after a part has had xml created and added to model, it is reusable

   # Create duplicate object
   dummy_2 = ReusableDummyObj(model = tmp_model)
   assert not bu.matching_uuids(dummy_1.uuid, dummy_2.uuid)
   assert dummy_2.try_reuse()  # should have matched dummy_1
   assert bu.matching_uuids(dummy_1.uuid, dummy_2.uuid)

   # Only one 'RESQML' object should exist in model, despite us having two resqpy objects
   uuids = tmp_model.uuids(obj_type = ReusableDummyObj.resqml_type)
   assert len(uuids) == 1
   assert bu.matching_uuids(uuids[0], dummy_2.uuid)



"""relperm.py: class for storing and retrieving dataframes of relative permeability
data as RESQML objects.

   note that this module uses the obj_Grid2dRepresentation class in a way that was not envisaged
   when the RESQML standard was defined; software that does not use resqpy is unlikely to be
   able to do much with data stored in this way

   Nexus is a registered trademark of Halliburton
"""
import numpy as np
import pandas as pd
import os
import logging
import resqpy.olio.xml_et as rqet
from resqpy.olio.dataframe import DataFrame

version = '27th July 2021'

log = logging.getLogger(__name__)
log.debug(f'dataframe.py version {version}')


class RelPerm(DataFrame):
   """Class for storing and retrieving a pandas dataframe of relative permeability data.

   note:
      inherits from DataFrame class
   """

   def __init__(self,
                model,
                uuid = None,
                df = None,
                uom_list = None,
                realization = None,
                phase_combo = None,
                low_sal = False,
                table_index = None,
                title = 'relperm_table',
                column_lookup_uuid = None,
                uom_lookup_uuid = None):
      """Create a new RelPerm object from either a previously stored object or a pandas dataframe.

      arguments:
         phase_combo (str, optional): the combination of phases whose relative permeability behaviour is described.
            Options include 'water-oil', 'gas-oil' and 'gas-water'
         low_sal (boolean, optional): if True, indicates that the water-oil table contains the low-salinity data for
            relative permeability and capillary pressure
         table_index (int, optional): the index of the relative permeability
         table when multiple relative permeability tables are present. Note, indices should start at 1.

      note:
         see DataFrame class docstring for details of other arguments
      """

      # check that either a uuid OR dataframe has been provided
      if df is None and uuid is None:
         raise ValueError('either a uuid or a dataframe must be provided')

      # check extra_metadata for arguments if uuid is provided
      if uuid is not None:
         model_root = model.root(uuid = uuid)
         uuid_metadata_dict = rqet.load_metadata_from_xml(model_root)
         phase_combo = uuid_metadata_dict['phase_combo']
         low_sal = uuid_metadata_dict['low_sal']
         table_index = uuid_metadata_dict['table_index']

      elif df is not None:
         # check that 'phase_combo' parameter is valid
         processed_phase_combo = set([x.strip() for x in str(phase_combo).split('-')])
         if processed_phase_combo not in [{'water', 'oil'}, {'gas', 'oil'}, {'gas', 'water'}, {'None'}]:
            raise ValueError('invalid phase_combo provided')
         # check that table_index is >= 1
         if table_index is not None:
            if table_index < 1:
               raise ValueError('table_index cannot be less than 1')
         # check that the column names and order are as expected
         df.columns = [x.capitalize() for x in df.columns]
         if 'Pc' in df.columns:
            if df.columns[-1] != 'Pc':
               raise ValueError('Pc', 'capillary pressure data should be in the last column of the dataframe')
         if phase_combo is not None:
            if processed_phase_combo == {'water', 'oil'}:
               expected_cols = {'Sw', 'So', 'Krw', 'Kro', 'Pc'}
               sat_cols = {'Sw', 'So'}
               if df.columns[0] not in sat_cols or len(set(df.columns).intersection(sat_cols)) != 1:
                  raise ValueError('incorrect saturation column name and/or multiple saturation columns exist')
               if not set(df.columns).issubset(expected_cols):
                  raise ValueError(f'incorrect column name(s) {set(df.columns).difference(expected_cols)}')
            elif processed_phase_combo == {'gas', 'oil'}:
               expected_cols = {'Sg', 'So', 'Krg', 'Kro', 'Pc'}
               sat_cols = {'Sg', 'So'}
               if df.columns[0] not in sat_cols or len(set(df.columns).intersection(sat_cols)) != 1:
                  raise ValueError('incorrect saturation column name and/or multiple saturation columns exist')
               if not set(df.columns).issubset(expected_cols):
                  raise ValueError(f'incorrect column name(s) {set(df.columns).difference(expected_cols)}')
            elif processed_phase_combo == {'gas', 'water'}:
               expected_cols = {'Sg', 'Sw', 'Krg', 'Krw', 'Pc'}
               sat_cols = {'Sg', 'Sw'}
               if df.columns[0] not in sat_cols or len(set(df.columns).intersection(sat_cols)) != 1:
                  raise ValueError('incorrect saturation column name and/or multiple saturation columns exist')
               if not set(df.columns).issubset(expected_cols):
                  raise ValueError(f'incorrect column name(s) {set(df.columns).difference(expected_cols)}')
         elif phase_combo is None:
            if df.columns[0] not in ['Sw', 'Sg', 'So'] or len(set(df.columns).intersection({'Sw', 'Sg', 'So'})) != 1:
               raise ValueError('incorrect saturation column name and/or multiple saturation columns exist')
            if set(df.columns).issubset({'Sw', 'So', 'Krw', 'Kro', 'Pc'}) and len(set(df.columns)) >= 3:
               phase_combo = 'water-oil'
            elif set(df.columns).issubset({'Sg', 'So', 'Krg', 'Kro', 'Pc'}) and len(set(df.columns)) >= 3:
               phase_combo = 'gas-oil'
            elif set(df.columns).issubset({'Sg', 'Sw', 'Krg', 'Krw', 'Pc'}) and len(set(df.columns)) >= 3:
               phase_combo = 'gas-water'
            else:
               raise Exception('unexpected number of columns and/or column headers')

         # ensure that missing capillary pressure values are stored as np.nan
         for col in df.columns:
            if col.capitalize() == 'Pc':
               df[col].replace('None', np.nan, inplace = True)

         # convert all values in the dataframe to numeric type
         df_cols = df.columns
         df[df_cols] = df[df_cols].apply(pd.to_numeric, errors = 'coerce')

         # ensure that no other column besides Pc has missing values
         cols_no_pc = [x for x in df.columns if 'pc' != x.lower()]
         for col in cols_no_pc:
            if (df[col].isnull().sum() > 0) | ('None' in list(df[col])):
               raise Exception(f'missing values found in {col} column')

         # check that Sw, Kr and Pc values are monotonic and that the Sw and Kr values are within the range 0-1
         sat_col = [x for x in df.columns if x[0] == 'S'][0]
         relp_corr_col = [x for x in df.columns if x == 'Kr' + sat_col[-1]][0]
         relp_opp_col = [x for x in df.columns if (x[0:2] == 'Kr') & (x[-1] != sat_col[-1])][0]
         if (df[sat_col].is_monotonic and df[relp_corr_col].is_monotonic and df[
             relp_opp_col].is_monotonic_decreasing) or \
             (df[sat_col].is_monotonic_decreasing and df[relp_corr_col].is_monotonic_decreasing and
                 df[relp_opp_col].is_monotonic):
            pass
         else:
            raise ValueError(f'{sat_col, relp_corr_col, relp_opp_col} combo is not monotonic')
         if 'Pc' in df.columns:
            if not df['Pc'].dropna().is_monotonic and not df['Pc'].dropna().is_monotonic_decreasing:
               raise ValueError('Pc values are not monotonic')
         for col in ['Sw', 'Sg', 'So', 'Krw', 'Krg', 'Kro']:
            if col in df.columns:
               if df[col].min() < 0 or df[col].max() > 1:
                  raise ValueError(f'{col} is not within the range 0-1')

      super().__init__(model,
                       uuid = uuid,
                       support_root = None,
                       df = df,
                       uom_list = uom_list,
                       realization = realization,
                       title = title,
                       column_lookup_uuid = column_lookup_uuid,
                       uom_lookup_uuid = uom_lookup_uuid,
                       extra_metadata = {
                          'relperm_table': 'true',
                          'phase_combo': phase_combo,
                          'low_sal': str(low_sal).lower(),
                          'table_index': str(table_index)
                       })
      self.phase_combo = phase_combo
      self.low_sal = low_sal
      self.table_index = table_index

   def interpolate_point(self, saturation, kr_or_pc_col):
      """Returns a tuple of the saturation value and the corresponding interpolated rel. perm. or cap. pressure value.

      arguments:
         saturation (float): the saturation at which the relative permeability or cap. pressure will be interpolated
         kr_or_pc_col (str): the column name of the parameter to be interpolated

      returns:
         tuple of float, the first element is the saturation and the second element is the interpolated value

      note:
         A simple linear interpolation is performed.
      """
      df = self.df.copy()
      if kr_or_pc_col.capitalize() not in df.columns or kr_or_pc_col.capitalize() == df.columns[0]:
         raise ValueError('incorrect column name provided for interpolation')
      if kr_or_pc_col == 'PC':
         df = df[df['PC'].notnull()]
      sat_col = df.columns[0]
      # ensure that the saturation values are monotonically increasing
      df = df.sort_values(by = sat_col)
      x = df[sat_col]
      y = df[kr_or_pc_col]
      x_new = saturation
      if x_new < x.min() or x_new > x.max():
         raise ValueError('saturation value is outside the interpolation range')
      y_new = np.interp(x_new, x, y)
      return saturation, y_new

   def df_to_text(self, filepath, filename):
      """Creates a text file from a dataframe of relative permeability and capillary pressure data.

      arguments:
         filepath (str): location where new text file is written to
         filename (str): name of the new text file

      returns:
         tuple of float, the first element is the saturation and the second element is the interpolated value

      note:
         Only Nexus compatible text files are currently supported. Text files that are compatible with other reservoir
            simulators may be supported in the future.
      """
      df = self.df.copy()
      ascii_file = os.path.join(filepath, filename + '.dat')
      df.columns = map(str.upper, df.columns)
      if {'KRW', 'KRO'}.issubset(set(df.columns)):
         if self.low_sal:
            table_name_keyword = 'WOTABLE (LOW_SAL)\n'
         else:
            table_name_keyword = 'WOTABLE\n'
         df.rename(columns = {'SW': 'SW', 'KRW': 'KRW', 'KRO': 'KROW', 'PC': 'PCWO'}, inplace = True)
      elif {'KRG', 'KRO'}.issubset(set(df.columns)):
         table_name_keyword = 'GOTABLE\n'
         df.rename(columns = {'SG': 'SG', 'KRG': 'KRG', 'KRO': 'KROG', 'PC': 'PCGO'}, inplace = True)
      elif {'KRW', 'KRW'}.issubset(set(df.columns)):
         table_name_keyword = 'GWTABLE\n'
         df.rename(columns = {'SG': 'SG', 'KRG': 'KRG', 'KRW': 'KRWG', 'PC': 'PCGW'}, inplace = True)
      else:
         raise Exception('incorrect rel. perm. column combination encountered')

      if filename + '.dat' not in os.listdir(filepath):
         with open(ascii_file, 'w') as f:
            f.write(table_name_keyword)
            df_str = df.to_string(na_rep = '', index = False)
            f.write(df_str)
            f.close()
            print(f'Created new DAT file: {filename} at {filepath}')
      else:
         with open(ascii_file, 'a') as f:
            f.write('\n\n')
            f.write(table_name_keyword)
            df_str = df.to_string(na_rep = '', index = False)
            f.write(df_str)
            f.close()
            print(f'Appended to DAT file: {filename} at {filepath}')

   def write_hdf5_and_create_xml(self):
      """Write relative permeability table data to hdf5 file and create xml for RESQML objects to represent dataframe.

      """
      super().write_hdf5_and_create_xml()
      mesh_root = self.mesh.root
      # create an xml of extra metadata to indicate that this is a relative permeability table
      rqet.create_metadata_xml(mesh_root, self.extra_metadata)


def text_to_relperm_dict(filepath):
   """Returns a dictionary that contains dataframes with relative permeability and capillary pressure data and phase combinations.

   arguments:
      filepath (str): relative or full path of the text file to be processed.

   returns:
      dict, each element in the dictionary contains a dataframe, with saturation and rel. permeability/capillary pressure
      data, and the phase combination being described

   note:
      Only Nexus compatible text files are currently supported. Text files from other reservoir simulators may be
         supported in the future.
   """
   with open(filepath) as f:
      # create list of rows of the original ascii file with blank lines removed
      data = list(filter(None, [list(filter(None, x.strip('\n').split(' '))) for x in f.readlines()]))
      # remove comments
      data = [x for x in data if ('!' not in x)]

   # get indices of start of each new relperm table based on Nexus keywords
   table_start_positions = [
      i for i, l in enumerate(data) if len({'WOTABLE', 'GOTABLE', 'GWTABLE'}.intersection(set(l))) == 1
   ]
   df_cols_dict = {
      'SW': 'Sw',
      'SG': 'Sg',
      'KRW': 'Krw',
      'KRG': 'Krg',
      'KROW': 'Kro',
      'KROG': 'Kro',
      'KRWG': 'Krw',
      'PCWO': 'Pc',
      'PCGO': 'Pc',
      'PCGW': 'Pc'
   }
   relperm_table_idx = 1
   rel_perm_dict = {}
   for i, l in enumerate(table_start_positions):
      key = 'relperm_table' + str(relperm_table_idx)
      rel_perm_dict[key] = {}
      relperm_table_idx += 1
      if 'WOTABLE' in data[l]:
         phase_combo = 'water-oil'
         rel_perm_dict[key]['phase_combo'] = phase_combo
      elif 'GOTABLE' in data[l]:
         phase_combo = 'gas-oil'
         rel_perm_dict[key]['phase_combo'] = phase_combo
      elif 'GWTABLE' in data[l]:
         phase_combo = 'gas-water'
         rel_perm_dict[key]['phase_combo'] = phase_combo
      else:
         raise Exception('incorrect table key word encountered')
      if i < (len(table_start_positions) - 1):
         table_end = table_start_positions[i + 1]
      else:
         table_end = len(data)
      table_cols = data[l + 1]
      table_rows = data[l + 2:table_end]
      df = pd.DataFrame(table_rows, columns = table_cols)
      df.columns = df.columns.map(df_cols_dict)
      sat_col = [x for x in df.columns if 'S' in x][0]
      df = df.apply(pd.to_numeric, errors = 'coerce')
      df = df[df[sat_col].notnull()]
      rel_perm_dict[key]['df'] = df
   return rel_perm_dict


def relperm_parts_in_model(model,
                           phase_combo = None,
                           low_sal = None,
                           table_index = None,
                           title = None,
                           related_uuid = None):
   """Returns list of part names within model that are representing RelPerm dataframe support objects.

   arguments:
      model (model.Model): the model to be inspected for dataframes
      phase_combo (str, optional): the combination of phases whose relative permeability behaviour is described.
         Options include 'water-oil', 'gas-oil', 'gas-water', 'oil-water', 'oil-gas' and 'water-gas'
      low_sal (boolean, optional): if True, indicates that the water-oil table contains the low-salinity data for
         relative permeability and capillary pressure
      table_index (int, optional): the index of the relative permeability table when multiple relative permeability
         tables are present. Note, indices should start at 1.
      title (str, optional): if present, only parts with a citation title exactly matching will be included
      related_uuid (uuid, optional): if present, only parts relating to this uuid are included

   returns:
      list of str, each element in the list is a part name, within model, which is representing the support for a RelPerm object
   """
   extra_metadata_orig = {
      'relperm_table': 'true',
      'phase_combo': phase_combo,
      'low_sal': low_sal,
      'table_index': table_index
   }
   extra_metadata = {k: str(v).lower() for k, v in extra_metadata_orig.items() if v is not None}
   df_parts_list = model.parts(obj_type = 'Grid2dRepresentation',
                               title = title,
                               extra = extra_metadata,
                               related_uuid = related_uuid)

   return df_parts_list



import pytest

import resqpy.organize as rqo
from resqpy.model import Model


# Test saving and loading from disk
@pytest.mark.parametrize(
   "cls, data",
   [
      (
         rqo.OrganizationFeature,
         dict(feature_name = 'hello', organization_kind = 'stratigraphic'),
      ),
      (
         rqo.GeobodyFeature,
         dict(feature_name = 'hi'),
      ),
      (
         rqo.BoundaryFeature,
         dict(feature_name = 'foobar'),
      ),
      (
         rqo.FrontierFeature,
         dict(feature_name = 'foobar'),
      ),
      (
         rqo.GeologicUnitFeature,
         dict(feature_name = 'foobar'),
      ),
      (
         rqo.FluidBoundaryFeature,
         dict(feature_name = 'foobar', kind = 'gas oil contact'),
      ),
      (
         rqo.TectonicBoundaryFeature,
         dict(feature_name = 'foobar', kind = 'fracture'),
      ),
      (
         rqo.GeneticBoundaryFeature,
         dict(feature_name = 'foobar', kind = 'geobody boundary'),  # Age as well?
      ),
      (
         rqo.WellboreFeature,
         dict(feature_name = 'foobar'),
      ),
   ])
def test_organize_classes(tmp_model, cls, data):

   # Load example model from a fixture
   model = tmp_model
   epc = model.epc_file

   # Create the feature
   obj = cls(parent_model = model, **data)
   uuid = obj.uuid

   # Save to disk
   obj.create_xml()
   model.store_epc()
   model.h5_release()

   # Reload from disk
   del model, obj
   model2 = Model(epc_file = epc)
   obj2 = cls(parent_model = model2, uuid = uuid)

   # Check all attributes were loaded correctly
   for key, expected_value in data.items():
      assert getattr(obj2, key) == expected_value, f"Error for {key}"


def test_RockFluidUnitFeature(tmp_model):

   # Create the features
   top = rqo.BoundaryFeature(tmp_model, feature_name = 'the top')
   base = rqo.BoundaryFeature(tmp_model, feature_name = 'the base')
   rfuf_1 = rqo.RockFluidUnitFeature(parent_model = tmp_model,
                                     feature_name = 'foobar',
                                     phase = 'seal',
                                     top_boundary_feature = top,
                                     base_boundary_feature = base)
   uuid = rfuf_1.uuid

   # Save to disk
   top.create_xml()
   base.create_xml()
   rfuf_1.create_xml()
   tmp_model.store_epc()

   # Reload from disk
   rfuf_2 = rqo.RockFluidUnitFeature(parent_model = tmp_model, uuid = uuid)

   # Check properties the same
   assert rfuf_2.feature_name == 'foobar'
   assert rfuf_2.phase == 'seal'
   assert rfuf_2.top_boundary_feature.feature_name == 'the top'
   assert rfuf_2.base_boundary_feature.feature_name == 'the base'


def test_FaultInterp(tmp_model):

   title = "not my fault"
   tect_boundary = rqo.TectonicBoundaryFeature(tmp_model, kind = 'fault')
   fault_interp = rqo.FaultInterpretation(
      tmp_model,
      tectonic_boundary_feature = tect_boundary,
      title = title,
      domain = "depth",
      is_normal = True,
      maximum_throw = 3,
   )

   tect_boundary.create_xml()
   fault_interp.create_xml()

   fault_interp_2 = rqo.FaultInterpretation(tmp_model, uuid = fault_interp.uuid)
   assert fault_interp_2.title == title
   assert fault_interp_2.maximum_throw == 3


def test_EarthModel(tmp_model):

   title = 'gaia'
   org_feat = rqo.OrganizationFeature(tmp_model, feature_name = 'marie kondo', organization_kind = "earth model")
   em1 = rqo.EarthModelInterpretation(tmp_model, title = title, organization_feature = org_feat)

   org_feat.create_xml()
   em1.create_xml()

   em2 = rqo.EarthModelInterpretation(tmp_model, uuid = em1.uuid)
   assert em2.title == title


def test_Horizon(tmp_model):

   gen = rqo.GeneticBoundaryFeature(tmp_model, kind = 'horizon')
   hor = rqo.HorizonInterpretation(tmp_model,
                                   genetic_boundary_feature = gen,
                                   sequence_stratigraphy_surface = 'maximum flooding')

   gen.create_xml()
   hor.create_xml()

   hor2 = rqo.HorizonInterpretation(tmp_model, uuid = hor.uuid)
   assert hor2.sequence_stratigraphy_surface == hor.sequence_stratigraphy_surface


def test_GeobodyBoundary(tmp_model):

   gen = rqo.GeneticBoundaryFeature(tmp_model, kind = 'geobody boundary')
   gb = rqo.GeobodyBoundaryInterpretation(tmp_model, genetic_boundary_feature = gen)

   gen.create_xml()
   gb.create_xml()
   gb2 = rqo.GeobodyBoundaryInterpretation(tmp_model, uuid = gb.uuid)
   assert gb == gb2


def test_wellbore_interp_title(tmp_model):
   # Create a feature and interp objects
   feature_name = 'well A'
   well_feature = rqo.WellboreFeature(tmp_model, feature_name = feature_name)
   well_feature.create_xml()
   well_interp_1 = rqo.WellboreInterpretation(tmp_model, wellbore_feature = well_feature, is_drilled = True)
   well_interp_1.create_xml()

   # Create a duplicate object, loading from XML
   well_interp_2 = rqo.WellboreInterpretation(tmp_model, uuid = well_interp_1.uuid)

   # Check feature name is present
   assert well_interp_1.title == feature_name
   assert well_interp_2.title == feature_name



"""Support for consolidation of datasets based on equivalence between parts ."""

version = '18th June 2021'

import logging

log = logging.getLogger(__name__)
log.debug('consolidation.py version ' + version)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet

import resqpy.crs as rqc
import resqpy.organize as rqo
import resqpy.property as rqp
import resqpy.time_series as rqt

# the following list contains those RESQML classes with equivalence methods in their resqpy class
# it is lightly ordered with earlier classes having no dependence on classes later in the list

consolidatable_list = [
   'OrganizationFeature', 'GeobodyFeature', 'BoundaryFeature', 'FrontierFeature', 'GeologicUnitFeature',
   'FluidBoundaryFeature', 'RockFluidUnitFeature', 'TectonicBoundaryFeature', 'GeneticBoundaryFeature',
   'WellboreFeature', 'FaultInterpretation', 'EarthModelInterpretation', 'HorizonInterpretation',
   'GeobodyBoundaryInterpretation', 'GeobodyInterpretation', 'WellboreInterpretation', 'LocalDepth3dCrs',
   'LocalTime3dCrs', 'TimeSeries', 'StringTableLookup', 'PropertyKind'
]

# todo: add to this list as other classes gain an is_equivalent() method


class Consolidation:

   def __init__(self, resident_model):

      self.model = resident_model
      log.debug(f'new consolidation for {self.model.epc_file} with {len(self.model.uuids())} uuids')
      self.map = {}  # dictionary mapping immigrant uuid to primary uuid
      self.stale = True

   def equivalent_uuid_for_part(self, part, immigrant_model = None, ignore_identical_part = False):
      """Returns uuid of an equivalent part in primary model, or None if no equivalent found."""

      #      log.debug('Looking for equivalent uuid for: ' + str(part))
      if not part:
         return None
      if immigrant_model is None:
         immigrant_model = self.model
      immigrant_uuid = rqet.uuid_in_part_name(part)
      #      log.debug('   immigrant uuid: ' + str(immigrant_uuid))
      if immigrant_uuid in self.map:
         #         log.debug('   known to be equivalent to: ' + str(self.map[immigrant_uuid]))
         return self.map[immigrant_uuid]
      obj_type = immigrant_model.type_of_part(part, strip_obj = True)
      if obj_type is None or obj_type not in consolidatable_list:
         return None
      #      log.debug('   object type is consolidatable')
      resident_uuids = self.model.uuids(obj_type = obj_type)
      if resident_uuids is None or len(resident_uuids) == 0:
         #         log.debug('   no resident parts found of type: ' + str(obj_type))
         return None
#      log.debug(f'   {len(resident_uuids)} resident parts of same class')
      if not ignore_identical_part:
         for resident_uuid in resident_uuids:
            if bu.matching_uuids(resident_uuid, immigrant_uuid):
               #               log.debug('   uuid already resident: ' + str(resident_uuid))
               return resident_uuid


#      log.debug('   preparing immigrant object')
      if obj_type.endswith('Interpretation') or obj_type.endswith('Feature'):
         immigrant_obj = rqo.__dict__[obj_type](immigrant_model, uuid = immigrant_uuid)
      elif obj_type.endswith('Crs'):
         immigrant_obj = rqc.Crs(immigrant_model, uuid = immigrant_uuid)
      elif obj_type == 'TimeSeries':
         immigrant_obj = rqt.TimeSeries(immigrant_model, uuid = immigrant_uuid)
      elif obj_type == 'StringTableLookup':
         immigrant_obj = rqp.StringLookup(immigrant_model, uuid = immigrant_uuid)
      elif obj_type == 'PropertyKind':
         immigrant_obj = rqp.PropertyKind(immigrant_model, uuid = immigrant_uuid)
      else:
         raise Exception('code failure')
      assert immigrant_obj is not None
      for resident_uuid in resident_uuids:
         #         log.debug('   considering resident: ' + str(resident_uuid))
         if ignore_identical_part and bu.matching_uuids(resident_uuid, immigrant_uuid):
            continue
         if obj_type.endswith('Interpretation') or obj_type.endswith('Feature'):
            resident_obj = rqo.__dict__[obj_type](self.model, uuid = resident_uuid)
         elif obj_type.endswith('Crs'):
            resident_obj = rqc.Crs(self.model, uuid = resident_uuid)
         elif obj_type == 'TimeSeries':
            resident_obj = rqt.TimeSeries(self.model, uuid = resident_uuid)
         elif obj_type == 'StringTableLookup':
            resident_obj = rqp.StringLookup(self.model, uuid = resident_uuid)
         elif obj_type == 'PropertyKind':
            resident_obj = rqp.PropertyKind(self.model, uuid = resident_uuid)
         else:
            raise Exception('code failure')
         assert resident_obj is not None
         #         log.debug('   comparing with: ' + str(resident_obj.uuid))
         if immigrant_obj == resident_obj:  # note: == operator overloaded with equivalence method for these classes
            while resident_uuid in self.map:
               #               log.debug('   following equivalence for: ' + str(resident_uuid))
               resident_uuid = self.map[resident_uuid]
            self.map[immigrant_uuid] = resident_uuid
            #            log.debug('   new equivalence found with: ' + str(resident_uuid))
            return resident_uuid
      return None

   def force_uuid_equivalence(self, immigrant_uuid, resident_uuid):
      """Forces immigrant object to be treated as equivalent to (same as) resident object, identified by uuids."""

      assert immigrant_uuid is not None and resident_uuid is not None
      if isinstance(immigrant_uuid, str):
         immigrant_uuid = bu.uuid_from_string(immigrant_uuid)
      if isinstance(resident_uuid, str):
         resident_uuid = bu.uuid_from_string(resident_uuid)
      if bu.matching_uuids(immigrant_uuid, resident_uuid):
         return
      assert immigrant_uuid not in self.map.values()

      self.map[immigrant_uuid] = resident_uuid

   def force_part_equivalence(self, immigrant_part, resident_part):
      """Forces immigrant part to be treated as equivalent to resident part."""

      assert immigrant_part is not None and resident_part is not None
      if immigrant_part == resident_part:
         return
      self.force_uuid_equivalence(rqet.uuid_in_part_name(immigrant_part), rqet.uuid_in_part_name(resident_part))

   def check_map_integrity(self):
      """Raises assertion failure if map contains any potentially circular references."""

      for immigrant in self.map.keys():
         assert immigrant not in self.map.values()
      for resident in self.map.values():
         assert resident not in self.map.keys()


def sort_parts_list(model, parts_list):
   """Returns a copy of the parts list sorted into the preferred order for consolidating."""

   def ordering(obj_type):
      if obj_type in consolidatable_list:
         return consolidatable_list.index(obj_type)
      seq = len(consolidatable_list)
      if obj_type.endswith('Interpretation'):
         seq += 1
      elif obj_type.endswith('Representation'):
         seq += 2
      elif obj_type.endswith('Property'):
         seq += 3
      return seq

   pair_list = [(ordering(model.type_of_part(part, strip_obj = True)), part) for part in parts_list]
   pair_list.sort()
   return [part for _, part in pair_list]



"""Units of measure module."""

from pathlib import Path
import json
import warnings
from functools import lru_cache
from resqpy.olio.exceptions import InvalidUnitError, IncompatibleUnitsError

version = '6th July 2021'

# physical constants
feet_to_metres = 0.3048
metres_to_feet = 1.0 / feet_to_metres
ft3_to_m3 = 0.028316846592
m3_to_ft3 = 1.0 / ft3_to_m3
bbl_to_m3 = 0.158987294928
m3_to_bbl = 1.0 / bbl_to_m3
psi_to_kPa = 44.482216152605 / 6.4516
kPa_to_psi = 1.0 / psi_to_kPa
d_to_s = float(24 * 60 * 60)
s_to_d = 1.0 / d_to_s

# Mapping from uom to set of common case-insensitive aliases
# Nb. No need to write out fractional combinations such as "bbl/day"
UOM_ALIASES = {
   # Lengths
   'm': {'m', 'metre', 'metres', 'meter', 'meters'},
   'ft': {'ft', 'foot', 'feet'},

   # Times
   'ms': {'ms', 'msec', 'millisecs', 'millisecond', 'milliseconds'},
   's': {'s', 'sec', 'secs', 'second', 'seconds'},
   'min': {'min', 'mins', 'minute', 'minutes'},
   'h': {'h', 'hr', 'hour', 'hours'},
   'd': {'day', 'days'},
   'wk': {'wk', 'week', 'weeks'},
   'a': {'a', 'yr', 'year', 'years'},

   # Ratios
   '%': {'%', 'pu', 'p.u.'},
   'm3/m3': {'m3/m3', 'v/v'},
   'g/cm3': {'g/cm3', 'g/cc'},

   # Volumes
   'bbl': {'bbl', 'stb'},
   '1000 bbl': {'1000 bbl', 'mstb', 'mbbl'},
   '1E6 bbl': {'1E6 bbl', 'mmstb', 'mmbbl'},
   '1E6 ft3': {'1E6 ft3', 'mmscf'},
   '1000 ft3': {'1000 ft3', 'mscf'},
   'm3': {'m3', 'sm3'},
   'ft3': {'ft3', 'scf'},

   # Other
   'gAPI': {'gapi'},
   'S': {'mho'},
   'mS': {'mmho'},
   'psi': {'psi', 'psia'},
   'Euc': {'count'},
}
# Mapping from alias to valid uom
UOM_ALIAS_MAP = {alias.casefold(): uom for uom, aliases in UOM_ALIASES.items() for alias in aliases}

# Set of uoms that can be safely matched case-insensitive
CASE_INSENSITIVE_UOMS = {'m', 'ft', 'm3', 'ft3', 'm3/m3', 'ft3/ft3', 'bbl', 'bar', 'psi', 'm3/d', 'bbl/d'}


@lru_cache(None)
def rq_uom(units, quantity = None):
   """Returns RESQML uom string equivalent to units
   
   Args:
      units (str): unit to coerce
      quantity (str, optional): if given, raise an exception if the uom is not supported
         for this quantity

   Returns:
      str: unit of measure

   Raises:
      InvalidUnitError: if units cannot be coerced into RESQML units for the given quantity
   """
   if not units:
      raise InvalidUnitError("Must provide non-empty unit")

   uom = _try_parse_unit(units)

   # May be a fraction: match each part against known aliases
   if uom is None and '/' in units:
      parts = units.split('/', 1)
      newpart0 = _try_parse_unit(parts[0])
      newpart1 = _try_parse_unit(parts[1])
      uom = _try_parse_unit(f"{newpart0}/{newpart1}")

   if uom is None:
      raise InvalidUnitError(f"Cannot coerce {units} into a valid RESQML unit of measure.")

   if quantity is not None:
      supported_uoms = valid_uoms(quantity = quantity)
      if uom not in supported_uoms:
         raise InvalidUnitError(f"Unit {uom} is not supported for quantity {quantity}.\n"
                                f"Supported units:\n{supported_uoms}")
   return uom


def convert(x, unit_from, unit_to, quantity = None, inplace = False):
   """Convert value between two compatible units

   Args:
      x (numeric or np.array): value(s) to convert
      unit_from (str): resqml uom
      unit_to (str): resqml uom
      quantity (str, optional): If provided, raise an exception if units are not supported
         by this quantity
      inplace (bool): if True, convert arrays in-place. Else, return new value

   Returns:
      Converted value(s)

   Raises:
      InvalidUnitError: if units cannot be coerced into RESQML units
      IncompatibleUnitsError: if units do not have compatible base units
   """

   # conversion data assume the formula "y=(A + Bx)/(C + Dx)" where "y" represents a value in the base unit.
   # Backwards formula: x=(A-Cy)/(Dy-B)
   # All current units have D==0

   uom1 = rq_uom(unit_from, quantity = quantity)
   uom2 = rq_uom(unit_to, quantity = quantity)

   if uom1 == uom2:
      return x

   base1, dim1, (A1, B1, C1, D1) = get_conversion_factors(uom1)
   base2, dim2, (A2, B2, C2, D2) = get_conversion_factors(uom2)

   if base1 != base2:
      if dim1 != dim2:
         raise IncompatibleUnitsError(f"Cannot convert from '{unit_from}' to '{unit_to}':"
                                      f"\n - '{uom1}' has base unit '{base1} and dimension '{dim1}'."
                                      f"\n - '{uom2}' has base unit '{base2} and dimension '{dim2}'.")
      else:
         warnings.warn(f"Assuming base units {base1} and {base2} are equivalent as they have the same dimensions:"
                       f"\n - '{uom1}' has base unit '{base1} and dimension '{dim1}'."
                       f"\n - '{uom2}' has base unit '{base2} and dimension '{dim2}'.")

   if not inplace:
      y = (A1 + (B1 * x)) / (C1 + (D1 * x))
      return (A2 - (C2 * y)) / ((D2 * y) - B2)

   else:
      if any(f != 0 for f in [A1, A2, D1, D2]):
         raise NotImplementedError("In-place conversion not yet implemented for non-trivial conversions")

      factor = (B1 * C2) / (C1 * B2)
      x *= factor
      return x


@lru_cache(None)
def valid_uoms(quantity = None, return_attributes = False):
   """Return set of valid RESQML units of measure
   
   Args:
      quantity (str): If given, filter to uoms supported by this quanitity.
      return_attributes (bool): If True, return a dict of all uoms and their
         attributes, such as the full name and dimension. Else, simply return
         the set of valid uoms.

   Returns
      set or dict
   """
   uoms = _properties_data()['units']
   if quantity:
      all_quantities = _properties_data()['quantities']
      supported_members = all_quantities[quantity]['members']
      uoms = {k: v for k, v in uoms.items() if k in supported_members}
   if return_attributes:
      return uoms
   else:
      return set(uoms.keys())


@lru_cache(None)
def valid_quantities(return_attributes = False):
   """Return set of valid RESQML quantities
   
   Args:
      return_attributes (bool): If True, return a dict of all quantities and their
         attributes, such as the supported units of measure. Else, simply return
         the set of valid properties.
   
   Returns
      set or dict
   """
   quantities = _properties_data()['quantities']
   if return_attributes:
      return quantities
   else:
      return set(quantities.keys())


def valid_property_kinds():
   """Return set of valid property kinds"""

   return set(_properties_data()['property_kinds'].keys())


def rq_uom_list(units_list):
   """Returns a list of RESQML uom equivalents for units in list."""

   return [rq_uom(u) for u in units_list]


def rq_length_unit(units):
   """Returns length units string as expected by resqml."""

   return rq_uom(units, quantity = 'length')


def rq_time_unit(units):
   """Returns time units string as expected by resqml."""

   return rq_uom(units, quantity = 'time')


def convert_times(a, from_units, to_units, invert = False):
   """Converts values in numpy array (or a scalar) from one time unit to another, in situ if array.
   
   note:
      To see supported units, use: `valid_uoms(quantity='time')`
   """

   if invert:
      from_units, to_units = to_units, from_units

   return convert(a, from_units, to_units, quantity = 'time', inplace = True)


def convert_lengths(a, from_units, to_units):
   """Converts values in numpy array (or a scalar) from one length unit to another, in situ if array.

   arguments:
      a (numpy float array, or float): array of length values to undergo unit conversion in situ, or a scalar
      from_units (string): the units of the data before conversion
      to_units (string): the required units

   returns:
      a after unit conversion

   note:
      To see supported units, use: `valid_uoms(quantity='length')`
   """

   return convert(a, from_units, to_units, quantity = 'length', inplace = True)


def convert_pressures(a, from_units, to_units):
   """Converts values in numpy array (or a scalar) from one pressure unit to another, in situ if array.

   arguments:
      a (numpy float array, or float): array of pressure values to undergo unit conversion in situ, or a scalar
      from_units (string): the units of the data before conversion
      to_units (string): the required units

   returns:
      a after unit conversion

   note:
      To see supported units, use: `valid_uoms(quantity='pressure')`
   """
   return convert(a, from_units, to_units, quantity = 'pressure', inplace = True)


def convert_volumes(a, from_units, to_units):
   """Converts values in numpy array (or a scalar) from one volume unit to another, in situ if array.

   arguments:
      a (numpy float array, or float): array of volume values to undergo unit conversion in situ, or a scalar
      from_units (string): units of the data before conversion; see note for accepted units
      to_units (string): the required units; see note for accepted units

   returns:
      a after unit conversion

   note:
      To see supported units, use: `valid_uoms(quantity='volume')`

   """
   return convert(a, from_units, to_units, quantity = 'volume', inplace = True)


def convert_flow_rates(a, from_units, to_units):
   """Converts values in numpy array (or a scalar) from one volume flow rate unit to another, in situ if array.

   arguments:
      a (numpy float array, or float): array of volume flow rate values to undergo unit conversion in situ, or a scalar
      from_units (string): units of the data before conversion, eg. 'm3/d'; see notes for acceptable units
      to_units (string): required units of the data after conversion, eg. 'ft3/d'; see notes for acceptable units

   returns:
      a after unit conversion

   note:
      To see supported units, use: `valid_uoms(quantity='volume per time')`
   """
   return convert(a, from_units, to_units, quantity = 'volume per time', inplace = True)


@lru_cache(None)
def get_conversion_factors(uom):
   """Return base unit and conversion factors (A, B, C, D) for a given uom.
   
   The formula "y=(A + Bx)/(C + Dx)" where "y" represents a value in the base unit.

   Returns:
      3-tuple of (base_unit, dimension, factors). Factors is a 4-tuple of conversion factors

   Raises:
      ValueError if either uom is not a valid resqml uom
   """
   if uom not in valid_uoms():
      raise ValueError(f"{uom} is not a valid uom")
   uoms_data = _properties_data()["units"][uom]

   dimension = uoms_data["dimension"]
   try:
      a, b, c, d = uoms_data["A"], uoms_data["B"], uoms_data["C"], uoms_data["D"]
      base_unit = uoms_data["baseUnit"]
   except KeyError:  # Base units do not have factors defined
      a, b, c, d = 0, 1, 1, 0
      base_unit = uom
   return base_unit, dimension, (a, b, c, d)


# Private functions


@lru_cache(None)
def _properties_data():
   """ Return a data structure that represents resqml unit system.

   The dict is loaded directly from a JSON file which is bundled with resqpy.
   The unit system is represented as a dict with the following keys:

   - dimensions
   - quantities
   - units
   - prefixes
   - property_kinds

   Returns:
      dict: resqml unit system
      
   """
   json_path = Path(__file__).parent / 'olio/data/properties.json'
   with open(json_path) as f:
      data = json.load(f)
   return data


def _try_parse_unit(units):
   """Try to match unit against known uoms and aliases, else return None"""

   uom_list = valid_uoms()
   ul = units.casefold()

   if units in uom_list:
      uom = units
   elif ul in CASE_INSENSITIVE_UOMS:
      uom = ul
   elif ul in UOM_ALIAS_MAP:
      uom = UOM_ALIAS_MAP[ul]
   elif ul in uom_list:
      uom = ul  # dangerous! for example, 'D' means D'Arcy and 'd' means day
   else:
      uom = None
   return uom



""" Script to construct JSON database of UOMs and property kinds.

Expects two files to exist in the same folder:
  - Energistics_Unit_of_Measure_Dictionary_V1.0.xml    (from units of measure standard)
  - Properties.xsd                                     (from resqml standard)

Creates properties.json containing parsed uoms and property kinds.
"""

from pathlib import Path
from lxml import etree
import json
import math as maths


def main():

   uom_dict_xml_path = Path(__file__).parent / 'Energistics_Unit_of_Measure_Dictionary_V1.0.xml'
   properties_xsd_path = Path(__file__).parent / 'Properties.xsd'
   json_path = Path(__file__).parent / 'properties.json'

   # Load xml
   assert uom_dict_xml_path.exists()
   assert properties_xsd_path.exists()
   uom_dict_root = etree.parse(str(uom_dict_xml_path)).getroot()
   properties_xsd_root = etree.parse(str(properties_xsd_path)).getroot()

   # Parse into memory
   data = dict(dimensions = parse_dimensions(uom_dict_root),
               quantities = parse_quantities(uom_dict_root),
               units = parse_units(uom_dict_root),
               prefixes = parse_prefixes(uom_dict_root),
               property_kinds = parse_property_kinds(properties_xsd_root))

   # Save to JSON
   with open(json_path, 'w', encoding = 'utf-8') as f:
      json.dump(data, f, ensure_ascii = False, indent = 2)


def parse_dimensions(root):

   dims = {}
   for node in root.find("{*}unitDimensionSet"):

      dimension = node.find('{*}dimension').text
      dims[dimension] = dict(
         name = node.find('{*}name').text,
         baseForConversion = node.find('{*}baseForConversion').text,
         canonicalUnit = node.find('{*}canonicalUnit').text,
      )
   return dims


def parse_quantities(root):

   quantities = {}
   for node in root.find("{*}quantityClassSet"):
      name = node.find('{*}name').text
      quantities[name] = dict(
         dimension = node.find('{*}dimension').text,
         baseForConversion = node.find('{*}baseForConversion').text,
      )
      alt = node.find('{*}alternativeBase')
      if alt is not None:
         quantities[name]['alternativeBase'] = alt.text

      members = []
      for member in node.findall('{*}memberUnit'):
         members.append(member.text)
      if members:
         quantities[name]["members"] = members

   return quantities


def parse_units(root):

   units = {}
   for node in root.find("{*}unitSet"):
      symbol = node.find('{*}symbol').text
      unit_dict = {}
      for key, typ in [
         ('name', str),
         ('dimension', str),
         ('isSI', as_bool),
         ('category', str),
         ('baseUnit', str),
         ('conversionRef', str),
         ('isExact', as_bool),
         ('A', as_numeric),
         ('B', as_numeric),
         ('C', as_numeric),
         ('D', as_numeric),
         ('conversionRef', str),
         ('description', str),
      ]:
         prop = node.find("{*}" + key)
         if prop is not None:
            unit_dict[key] = typ(prop.text)

      units[symbol] = unit_dict

   return units


def parse_prefixes(root):
   prefixes = {}
   for node in root.find("{*}prefixSet"):
      name = node.find('{*}name').text
      prefixes[name] = dict(
         symbol = node.find('{*}symbol').text,
         multiplier = node.find('{*}multiplier').text,
      )
      common_name = node.find('{*}commonName')
      if common_name is not None:
         prefixes[name]['common_name'] = common_name.text

   return prefixes


def parse_property_kinds(root):
   """ Return dict of valid RESQML property kinds

   Dict keys are the valid property kinds, e.g. 'angle per time'
   Dict values are the description, which may be None
   """

   kind_list = get_nodes_with_name(root, 'ResqmlPropertyKind')[1]
   kind_dict = {}
   for child in kind_list:
      kind = child.get('value')
      try:
         doc = child[0][0].text
      except IndexError:
         doc = None
      kind_dict[kind] = doc
   return kind_dict


def get_nodes_with_name(root, name):
   """ Find xml child node with given name """
   for child in root:
      if child.get('name') == name:
         return child
   raise ValueError(f'{name} not found')


def as_bool(a_string):
   return a_string.lower() == "true"


def as_numeric(a_string):
   if a_string == 'PI':
      return maths.pi
   elif a_string == "2*PI":
      return 2 * maths.pi
   elif a_string == "4*PI":
      return 4 * maths.pi
   elif a_string.isdigit():
      return int(a_string)
   else:
      return float(a_string)


if __name__ == "__main__":
   main()
